# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import pandas as pd
import pytest

from gluonts.model.forecast import (
    QuantileForecast,
    SampleForecast,
    _linear_interpolation,
    ExponentialTailApproximation,
)

QUANTILES = np.arange(1, 100) / 100
SAMPLES = np.arange(101).reshape(101, 1) / 100
FREQ = "1D"
START_DATE = pd.Period("2017 01-01 12:00", FREQ)

FORECASTS = {
    "QuantileForecast": QuantileForecast(
        forecast_arrays=QUANTILES.reshape(-1, 1),
        start_date=START_DATE,
        forecast_keys=np.array(QUANTILES, str),
    ),
    "SampleForecast": SampleForecast(
        samples=SAMPLES,
        start_date=START_DATE,
    ),
}

MULTIVARIATE_FORECASTS = {
    "SampleForecast": SampleForecast(
        samples=np.arange(160).reshape(8, 5, 4) / 100,
        start_date=START_DATE,
    ),
    "QuantileForecast": QuantileForecast(
        forecast_arrays=np.random.normal(size=(np.size(QUANTILES), 7, 3)),
        forecast_keys=np.array(QUANTILES, str),
        start_date=START_DATE,
    ),
}


@pytest.mark.parametrize("name", FORECASTS.keys())
def test_forecast(name):
    forecast = FORECASTS[name]

    def percentile(value):
        return f"p{int(round(value * 100)):02d}"

    num_samples, pred_length = SAMPLES.shape

    for quantile in QUANTILES:
        test_cases = [quantile, str(quantile), percentile(quantile)]
        for quant_pred in map(forecast.quantile, test_cases):
            assert np.isclose(
                quant_pred[0], quantile
            ), f"Expected {percentile(quantile)} quantile {quantile}. Obtained {quant_pred}."

    assert forecast.prediction_length == 1
    assert len(forecast.index) == pred_length
    assert forecast.index[0] == START_DATE

    forecast.plot()


def test_mean_only_forecast():
    forecast = QuantileForecast(
        forecast_arrays=np.ones(shape=(1, 12)),
        start_date=pd.Period("2022-03-04 00", freq="H"),
        forecast_keys=["mean"],
    )

    assert forecast.prediction_length == 12
    assert len(forecast.index) == 12
    assert forecast.index[0] == pd.Period("2022-03-04 00", freq="H")

    for level in [0.1, 0.5, 0.7]:
        assert np.isnan(forecast.quantile(level)).all()

    assert np.equal(forecast.mean, 1).all()


@pytest.mark.parametrize(
    "forecast, exp_index",
    [
        (
            SampleForecast(
                samples=np.random.normal(size=(100, 7, 3)),
                start_date=pd.Period("2020-01-01 00:00:00", freq="2D"),
            ),
            pd.period_range(
                start=pd.Period("2020-01-01 00:00:00", freq="2D"),
                periods=7,
                freq="2D",
            ),
        ),
    ],
)
def test_forecast_multivariate(forecast, exp_index):
    assert forecast.prediction_length == len(exp_index)
    assert np.all(forecast.index == exp_index)


@pytest.mark.parametrize("name", MULTIVARIATE_FORECASTS.keys())
def test_copy_dim(name):
    forecast = MULTIVARIATE_FORECASTS[name]
    for dim in range(forecast.dim()):
        univariate_forecast = forecast.copy_dim(dim)

        assert univariate_forecast.dim() == 1
        assert univariate_forecast.start_date == forecast.start_date
        assert univariate_forecast.item_id == forecast.item_id
        assert univariate_forecast.info == forecast.info

        if name == "SampleForecast":
            assert np.array_equal(
                univariate_forecast.samples,
                MULTIVARIATE_FORECASTS[name].samples[:, :, dim],
            )
        else:
            assert np.array_equal(
                univariate_forecast.forecast_array,
                MULTIVARIATE_FORECASTS[name].forecast_array[:, :, dim],
            )


def test_linear_interpolation() -> None:
    tol = 1e-7
    x_coord = [0.1, 0.5, 0.9]
    y_coord = [
        np.array([0.1, 0.5, 1]),
        np.array([1.0, 2.0, 3.0]),
        np.array([0.25, 0.5, 0.9]),
    ]

    x = 0.75
    exact = y_coord[1] + (x - x_coord[1]) * (y_coord[2] - y_coord[1]) / (
        x_coord[2] - x_coord[1]
    )
    assert np.all(
        np.abs(exact - _linear_interpolation(x_coord, y_coord, x)) <= tol
    )


def test_exponential_left_tail_approximation() -> None:
    tol = 1e-5
    x_coord = [0.1, 0.5, 0.9]
    y_coord = [
        np.array([0.1, 0.5, 1]),
        np.array([1.0, 2.0, 3.0]),
        np.array([0.25, 0.5, 0.9]),
    ]
    x = 0.01
    beta_inv = np.array([0.55920144, 0.9320024, 1.24266987])
    exact = beta_inv * np.log(x / x_coord[1]) + y_coord[1]
    exp_tail_approximation = ExponentialTailApproximation(x_coord, y_coord)
    assert np.all(np.abs(exact - exp_tail_approximation.left(x)) <= tol)


def test_exponential_right_tail_approximation() -> None:
    tol = 1e-5
    x_coord = [0.1, 0.5, 0.9]
    y_coord = [
        np.array([0.1, 0.5, 1]),
        np.array([1.0, 2.0, 3.0]),
        np.array([0.25, 0.5, 0.9]),
    ]
    x = 0.99
    beta_inv = np.array([-0.4660012, -0.9320024, -1.30480336])
    exact = beta_inv * np.log((1 - x_coord[1]) / (1 - x)) + y_coord[1]
    exp_tail_approximation = ExponentialTailApproximation(x_coord, y_coord)
    assert np.all(np.abs(exact - exp_tail_approximation.right(x)) <= tol)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from gluonts.ext.rotbaum import TreeEstimator

from gluonts.testutil.dummy_datasets import make_dummy_datasets_with_features

# TODO: Add support for categorical and dynamic features.


@pytest.mark.parametrize(
    "datasets",
    [
        # No features
        make_dummy_datasets_with_features(),
        # Single static categorical feature
        make_dummy_datasets_with_features(cardinality=[5]),
        # Multiple static categorical features
        make_dummy_datasets_with_features(cardinality=[3, 10, 42]),
        # Multiple static categorical features
        make_dummy_datasets_with_features(cardinality=[3, 10, 42]),
        # Single dynamic real feature
        make_dummy_datasets_with_features(num_feat_dynamic_real=1),
        # Multiple dynamic real feature
        make_dummy_datasets_with_features(num_feat_dynamic_real=3),
        # Multiple dynamic real feature
        make_dummy_datasets_with_features(num_feat_dynamic_real=3),
        # Both static categorical and dynamic real features
        make_dummy_datasets_with_features(
            cardinality=[3, 10, 42], num_feat_dynamic_real=3
        ),
        # Both static categorical and dynamic real features
        make_dummy_datasets_with_features(
            cardinality=[3, 10, 42], num_feat_dynamic_real=3
        ),
    ],
)
def test_rotbaum_smoke(datasets):
    dataset_train, dataset_test = datasets
    hps = {
        "freq": "D",
        "prediction_length": 3,
        "context_length": 3,
    }
    estimator = TreeEstimator.from_inputs(dataset_train, **hps)

    predictor = estimator.train(dataset_train)
    forecasts = list(predictor.predict(dataset_test))
    assert len(forecasts) == len(dataset_test)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import pandas as pd
import pytest

from gluonts.dataset.field_names import FieldName
from gluonts.dataset.split import (
    OffsetSplitter,
    periods_between,
    split,
    slice_data_entry,
)


def test_time_series_slice():
    entry = {
        "start": pd.Period("2021-02-03", "D"),
        "target": np.array(range(100), dtype=float),
        "feat_dynamic_real": np.expand_dims(
            np.array(range(100), dtype=float), 0
        ),
    }

    entry_slice = slice_data_entry(entry, slice(10, 20))
    assert entry_slice["start"] == pd.Period("2021-02-03", "D") + 10
    assert (entry_slice["target"] == np.arange(10, 20)).all()
    assert (
        entry_slice["feat_dynamic_real"] == np.array([np.arange(10, 20)])
    ).all()

    entry_slice = slice_data_entry(entry, slice(None, -20))
    assert entry_slice["start"] == pd.Period("2021-02-03", "D")
    assert (entry_slice["target"] == np.arange(80)).all()
    assert (
        entry_slice["feat_dynamic_real"] == np.array([np.arange(80)])
    ).all()

    entry_slice = slice_data_entry(entry, slice(-20, None))
    assert entry_slice["start"] == pd.Period("2021-02-03", "D") + 80
    assert (entry_slice["target"] == np.arange(80, 100)).all()
    assert (
        entry_slice["feat_dynamic_real"] == np.array([np.arange(80, 100)])
    ).all()


@pytest.mark.parametrize(
    "start, end, count",
    [
        (
            pd.Period("2021-03-04", freq="2D"),
            pd.Period("2021-03-05", freq="2D"),
            1,
        ),
        (
            pd.Period("2021-03-04", freq="2D"),
            pd.Period("2021-03-08", freq="2D"),
            3,
        ),
        (
            pd.Period("2021-03-03 23:00", freq="30T"),
            pd.Period("2021-03-04 03:29", freq="30T"),
            9,
        ),
        (
            pd.Period("2015-04-07 00:00", freq="30T"),
            pd.Period("2015-04-07 09:31", "30T"),
            20,
        ),
        (
            pd.Period("2015-04-07 00:00", freq="30T"),
            pd.Period("2015-04-08 16:10", freq="30T"),
            81,
        ),
        (
            pd.Period("2021-01-01 00", freq="2H"),
            pd.Period("2021-01-01 08", "2H"),
            5,
        ),
        (
            pd.Period("2021-01-01 00", freq="2H"),
            pd.Period("2021-01-01 11", "2H"),
            6,
        ),
        (
            pd.Period("2021-03-04", freq="2D"),
            pd.Period("2021-03-02", freq="2D"),
            0,
        ),
        (
            pd.Period("2021-03-04", freq="2D"),
            pd.Period("2021-03-04", freq="2D"),
            1,
        ),
        (
            pd.Period("2021-03-03 23:00", freq="30T"),
            pd.Period("2021-03-03 03:29", freq="30T"),
            0,
        ),
    ],
)
def test_periods_between(start, end, count):
    assert count == periods_between(start, end)


def test_negative_offset_splitter():
    dataset = [
        {
            "item_id": 0,
            "start": pd.Period("2021-03-04", freq="D"),
            "target": np.ones(shape=(100,)),
        },
        {
            "item_id": 1,
            "start": pd.Period("2021-03-04", freq="D"),
            "target": 2 * np.ones(shape=(50,)),
        },
    ]

    train, test_gen = OffsetSplitter(offset=-7).split(dataset)

    assert [len(t["target"]) for t in train] == [93, 43]
    assert [
        len(t["target"]) + len(s["target"])
        for t, s in test_gen.generate_instances(prediction_length=7)
    ] == [100, 50]

    train, test_gen = OffsetSplitter(offset=-21).split(dataset)

    assert [len(t["target"]) for t in train] == [79, 29]
    assert [
        len(t["target"]) + len(s["target"])
        for t, s in test_gen.generate_instances(prediction_length=7, windows=3)
    ] == [
        86,
        93,
        100,
        36,
        43,
        50,
    ]


def check_training_validation(
    original_entry,
    train_entry,
    valid_pair,
    prediction_length: int,
    max_history,
    date,
    offset,
) -> None:
    assert original_entry[FieldName.ITEM_ID] == train_entry[FieldName.ITEM_ID]
    assert train_entry[FieldName.ITEM_ID] == valid_pair[0][FieldName.ITEM_ID]
    if max_history is not None:
        assert valid_pair[0][FieldName.TARGET].shape[-1] == max_history
    assert valid_pair[1][FieldName.TARGET].shape[-1] == prediction_length
    train_end = (
        train_entry[FieldName.START]
        + train_entry[FieldName.TARGET].shape[-1]
        * train_entry[FieldName.START].freq
    )
    if date is not None:
        assert train_end == date + train_entry[FieldName.START].freq
    if offset is not None:
        if offset > 0:
            assert train_entry[FieldName.TARGET].shape[-1] == offset
        else:
            assert (
                train_entry[FieldName.TARGET].shape[-1] - offset
                == original_entry[FieldName.TARGET].shape[-1]
            )
    assert train_end <= valid_pair[1][FieldName.START]
    valid_end = (
        valid_pair[0][FieldName.START]
        + valid_pair[0][FieldName.TARGET].shape[-1]
        * valid_pair[0][FieldName.START].freq
    )
    assert valid_end == valid_pair[1][FieldName.START]
    if FieldName.FEAT_DYNAMIC_REAL in valid_pair[0]:
        assert (
            valid_pair[0][FieldName.FEAT_DYNAMIC_REAL].shape[-1]
            == valid_pair[0][FieldName.TARGET].shape[-1] + prediction_length
        )


@pytest.mark.parametrize(
    "dataset",
    [
        [
            {
                "item_id": 0,
                "start": pd.Period("2021-03-04", freq="D"),
                "target": np.ones(shape=(365,)),
                "feat_dynamic_real": 2 * np.ones(shape=(1, 365)),
            },
            {
                "item_id": 1,
                "start": pd.Period("2021-03-04", freq="D"),
                "target": 2 * np.ones(shape=(265,)),
                "feat_dynamic_real": 3 * np.ones(shape=(1, 265)),
            },
        ],
        [
            {
                "item_id": 0,
                "start": pd.Period("2021-03-04", freq="D"),
                "target": np.stack(
                    [np.ones(shape=(365,)), 10 * np.ones(shape=(365,))]
                ),
                "feat_dynamic_real": 2 * np.ones(shape=(1, 365)),
            },
            {
                "item_id": 1,
                "start": pd.Period("2021-03-04", freq="D"),
                "target": np.stack(
                    [2 * np.ones(shape=(265,)), 20 * np.ones(shape=(265,))]
                ),
                "feat_dynamic_real": 3 * np.ones(shape=(1, 265)),
            },
        ],
    ],
)
@pytest.mark.parametrize(
    "date, offset, windows, distance, max_history",
    [
        (pd.Period("2021-06-17", freq="D"), None, 1, None, None),
        (pd.Period("2021-06-17", freq="D"), None, 1, None, 96),
        (pd.Period("2021-06-17", freq="D"), None, 3, None, None),
        (pd.Period("2021-06-17", freq="D"), None, 3, None, 96),
        (None, 192, 1, None, 96),
        (None, 192, 3, None, 96),
        (None, -48, 1, None, None),
    ],
)
def test_split(dataset, date, offset, windows, distance, max_history):
    assert (offset is None) != (date is None)

    prediction_length = 7

    train, test_template = split(dataset, date=date, offset=offset)
    train = list(train)

    validation = list(
        test_template.generate_instances(
            prediction_length=prediction_length,
            windows=windows,
            distance=distance,
            max_history=max_history,
        )
    )

    assert len(train) == len(dataset)
    assert len(validation) == windows * len(train)

    k = 0
    for original_entry, train_entry in zip(dataset, train):
        for _ in range(windows):
            valid_pair = validation[k]
            check_training_validation(
                original_entry=original_entry,
                train_entry=train_entry,
                valid_pair=valid_pair,
                prediction_length=prediction_length,
                max_history=max_history,
                date=date,
                offset=offset,
            )
            k += 1


@pytest.mark.parametrize(
    "entry, offset, prediction_length, test_label_start",
    [
        (
            {
                "start": pd.Period("2015-04-07 00:00:00", freq="30T"),
                "target": np.random.randn(100),
            },
            20,
            6,
            pd.Period("2015-04-07 10:00:00", freq="30T"),
        ),
        (
            {
                "start": pd.Period("2015-04-07 00:00:00", freq="30T"),
                "target": np.random.randn(100),
            },
            -20,
            6,
            pd.Period("2015-04-08 16:00:00", freq="30T"),
        ),
    ],
)
def test_split_offset(
    entry,
    offset,
    prediction_length,
    test_label_start,
):
    training_dataset, test_template = split([entry], offset=offset)

    training_entry = next(iter(training_dataset))
    test_input, test_label = next(
        iter(
            test_template.generate_instances(
                prediction_length=prediction_length
            )
        )
    )

    if offset < 0:
        training_size = (len(entry["target"]) + offset,)
    else:
        training_size = (offset,)

    assert training_entry["start"] == entry["start"]
    assert training_entry["target"].shape == training_size

    assert test_input["start"] == entry["start"]
    assert test_input["target"].shape == training_size

    assert test_label["start"] == test_label_start
    assert test_label["target"].shape == (prediction_length,)


@pytest.mark.parametrize(
    "entry, date, prediction_length, training_size",
    [
        (
            {
                "start": pd.Period("2015-04-07 00:00:00", freq="30T"),
                "target": np.random.randn(100),
            },
            pd.Period("2015-04-07 09:30", "30T"),
            6,
            (20,),
        ),
        (
            {
                "start": pd.Period("2015-04-07 00:00:00", freq="30T"),
                "target": np.random.randn(100),
            },
            pd.Period("2015-04-08 16:00:00", freq="30T"),
            6,
            (81,),
        ),
        (
            {
                "start": pd.Period("2021-01-01 00", freq="2H"),
                "target": np.arange(10),
            },
            pd.Period("2021-01-01 08", "2h"),
            2,
            (5,),
        ),
        (
            {
                "start": pd.Period("2021-01-01 00", freq="2H"),
                "target": np.arange(10),
            },
            pd.Period("2021-01-01 11", "2h"),
            2,
            (6,),
        ),
    ],
)
def test_split_date(
    entry,
    date,
    prediction_length,
    training_size,
):
    training_dataset, test_template = split([entry], date=date)

    training_entry = next(iter(training_dataset))
    test_input, test_label = next(
        iter(
            test_template.generate_instances(
                prediction_length=prediction_length
            )
        )
    )

    assert training_entry["start"] == entry["start"]
    assert training_entry["target"].shape == training_size

    assert test_input["start"] == entry["start"]
    assert test_input["target"].shape == training_size

    assert test_label["start"] == test_input["start"] + len(
        test_input["target"]
    )
    assert test_label["target"].shape == (prediction_length,)


@pytest.mark.parametrize(
    "dataset",
    [
        [
            {
                "start": pd.Period("2021-03-01", freq="D"),
                "target": np.ones(shape=(28,)),
            }
        ],
    ],
)
@pytest.mark.parametrize(
    "date, offset, windows, distance",
    [
        (pd.Period("2021-03-22", freq="D"), None, 1, None),
        (pd.Period("2021-03-21", freq="D"), None, 2, 1),
        (pd.Period("2021-03-21", freq="D"), None, 2, 7),
        (None, 22, 1, None),
        (None, 21, 2, 1),
        (None, 21, 2, 7),
        (None, -6, 1, None),
        (None, -7, 2, 1),
        (None, -7, 2, 7),
    ],
)
def test_invalid_offset(dataset, date, offset, windows, distance):
    assert (offset is None) != (date is None)
    exp_msg = "Not enough data to generate some of the windows"
    prediction_length = 7

    _, test_template = split(dataset, date=date, offset=offset)
    with pytest.raises(AssertionError) as excinfo:
        list(
            test_template.generate_instances(
                prediction_length=prediction_length,
                windows=windows,
                distance=distance,
            )
        )
    assert exp_msg in str(excinfo.value)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from collections import UserDict
from dataclasses import dataclass
from operator import attrgetter
from typing import Any, Callable, Dict, Type, Tuple

from toolz import valmap


@dataclass
class Input:
    shape: Tuple[int, ...]
    dtype: Any
    required: bool = True


@dataclass
class InputSpec(UserDict):
    data: Dict[str, Input]
    zeros_fn: Callable

    @property
    def shapes(self) -> Dict[str, Tuple[int, ...]]:
        return valmap(attrgetter("shape"), self)

    @property
    def dtypes(self) -> Dict[str, Type]:
        return valmap(attrgetter("dtype"), self)

    def zeros(self):
        return {
            name: self.zeros_fn(input.shape, dtype=input.dtype)
            for name, input in self.items()
        }



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import pytest

from gluonts.ext.rotbaum import TreeEstimator


@pytest.fixture()
def hyperparameters(dsinfo):
    return dict(
        context_length=2,
        quantiles=[0.1, 0.5, 0.9],
        num_workers=0,
    )


@pytest.mark.parametrize("quantiles", [[0.1, 0.5, 0.9], [0.5]])
def test_accuracy(accuracy_test, hyperparameters, quantiles):
    hyperparameters.update(quantiles=quantiles, max_workers=32)

    accuracy_test(TreeEstimator, hyperparameters, accuracy=0.20)


def test_serialize(serialize_test, hyperparameters):
    serialize_test(TreeEstimator, hyperparameters)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx
import numpy as np
import pytest

from gluonts.core.serde import dump_json, load_json

from gluonts.mx.distribution import (
    Beta,
    Binned,
    Categorical,
    EmpiricalDistribution,
    Gamma,
    Gaussian,
    GenPareto,
    Laplace,
    NegativeBinomial,
    OneInflatedBeta,
    Poisson,
    StudentT,
    Uniform,
    ZeroAndOneInflatedBeta,
    ZeroInflatedBeta,
)

test_cases = [
    (
        Gaussian,
        {
            "mu": mx.nd.array([1000.0, -1000.0]),
            "sigma": mx.nd.array([0.1, 1.0]),
        },
    ),
    (
        Gamma,
        {"alpha": mx.nd.array([2.5, 7.0]), "beta": mx.nd.array([1.5, 2.1])},
    ),
    (
        GenPareto,
        {
            "xi": mx.nd.array([1 / 3.0, 1 / 4.0]),
            "beta": mx.nd.array([1.0, 2.0]),
        },
    ),
    (
        Beta,
        {"alpha": mx.nd.array([2.5, 7.0]), "beta": mx.nd.array([1.5, 2.1])},
    ),
    (
        Laplace,
        {"mu": mx.nd.array([1000.0, -1000.0]), "b": mx.nd.array([0.1, 1.0])},
    ),
    (
        StudentT,
        {
            "mu": mx.nd.array([1000.0, -1000.0]),
            "sigma": mx.nd.array([1.0, 2.0]),
            "nu": mx.nd.array([4.2, 3.0]),
        },
    ),
    (
        NegativeBinomial,
        {"mu": mx.nd.array([1000.0, 1.0]), "alpha": mx.nd.array([1.0, 2.0])},
    ),
    (
        Uniform,
        {
            "low": mx.nd.array([1000.0, -1000.1]),
            "high": mx.nd.array([2000.0, -1000.0]),
        },
    ),
    (
        Binned,
        {
            "bin_log_probs": mx.nd.array(
                [[1e-300, 0.3, 0.1, 0.05, 0.2, 0.1, 0.25]]
            )
            .log()
            .repeat(axis=0, repeats=2),
            "bin_centers": mx.nd.array(
                [[-5, -3, -1.2, -0.5, 0, 0.1, 0.2]]
            ).repeat(axis=0, repeats=2),
        },
    ),
    (
        Binned,
        {
            "bin_log_probs": mx.nd.array(
                [[1e-300, 0.3, 0.1, 0.05, 0.2, 0.1, 0.25]]
            )
            .log()
            .repeat(axis=0, repeats=2),
            "bin_centers": mx.nd.array(
                [[-5, -3, -1.2, -0.5, 0, 0.1, 0.2]]
            ).repeat(axis=0, repeats=2),
            "label_smoothing": 0.1,
        },
    ),
    (
        Categorical,
        {
            "log_probs": mx.nd.array(
                [[1e-300, 0.3, 0.1, 0.05, 0.2, 0.1, 0.25]]
            )
            .log()
            .repeat(axis=0, repeats=2),
        },
    ),
    (Poisson, {"rate": mx.nd.array([1000.0, 1.0])}),
    (
        ZeroInflatedBeta,
        {
            "alpha": mx.nd.array([0.175]),
            "beta": mx.nd.array([0.6]),
            "zero_probability": mx.nd.array([0.3]),
        },
    ),
    (
        OneInflatedBeta,
        {
            "alpha": mx.nd.array([0.175]),
            "beta": mx.nd.array([0.6]),
            "one_probability": mx.nd.array([0.3]),
        },
    ),
    (
        ZeroAndOneInflatedBeta,
        {
            "alpha": mx.nd.array([0.175]),
            "beta": mx.nd.array([0.6]),
            "zero_probability": mx.nd.array([0.3]),
            "one_probability": mx.nd.array([0.3]),
        },
    ),
    (
        EmpiricalDistribution,
        {
            "samples": mx.nd.stack(
                *[
                    mx.nd.arange(start=0, stop=20, step=2),
                    mx.nd.arange(start=100, stop=0, step=-10),
                ]
            ).transpose(),
            "event_dim": 1,
        },
    ),
]

test_output = {
    "Gaussian": {
        "mean": mx.nd.array([1000.0, -1000.0]),
        "stddev": mx.nd.array([0.1, 1.0]),
        "variance": mx.nd.array([0.01, 1.0]),
    },
    "Beta": {
        "mean": mx.nd.array([0.625, 0.7692307]),
        "stddev": mx.nd.array([0.2165063, 0.1325734]),
        "variance": mx.nd.array([0.046875, 0.0175757]),
    },
    "Gamma": {
        "mean": mx.nd.array([1.6666666, 3.3333333]),
        "stddev": mx.nd.array([1.05409255, 1.25988158]),
        "variance": mx.nd.array([1.1111111, 1.58730159]),
    },
    "GenPareto": {
        "mean": mx.nd.array([1.5, 2.666666666666666]),
        "stddev": mx.nd.array([2.5980762, 3.7712361663282534]),
        "variance": mx.nd.array([6.75, 14.222222222222221]),
    },
    "Laplace": {
        "mean": mx.nd.array([1000.0, -1000.0]),
        "stddev": mx.nd.array([0.14142136, 1.4142135]),
        "variance": mx.nd.array([0.02, 1.9999999]),
    },
    "StudentT": {
        "mean": mx.nd.array([1000.0, -1000.0]),
        "stddev": mx.nd.array([1.3816986, 3.4641016]),
        "variance": mx.nd.array([1.909091, 12.0]),
    },
    "NegativeBinomial": {
        "mean": mx.nd.array([1000.0, 1.0]),
        "stddev": mx.nd.array([1000.4999, 1.7320508]),
        "variance": mx.nd.array([1.001e06, 3.000e00]),
    },
    "Uniform": {
        "mean": mx.nd.array([1500.0, -1000.05]),
        "stddev": mx.nd.array([2.8867514e02, 2.8860467e-02]),
        "variance": mx.nd.array([8.3333336e04, 8.3292654e-04]),
    },
    "Binned": {
        "mean": mx.nd.array([-0.985, -0.985]),
        "stddev": mx.nd.array([1.377416, 1.377416]),
        "variance": mx.nd.array([1.8972749, 1.8972749]),
    },
    "Categorical": {
        "mean": mx.nd.array([3.45, 3.45]),
        "stddev": mx.nd.array([1.9868319, 1.9868319]),
        "variance": mx.nd.array([3.947501, 3.947501]),
    },
    "Poisson": {
        "mean": mx.nd.array([1000.0, 1.0]),
        "stddev": mx.nd.array([31.622776, 1.0]),
        "variance": mx.nd.array([1000.0, 1.0]),
    },
    "ZeroInflatedBeta": {
        "mean": mx.nd.array([0.15806451612903227]),
        "stddev": mx.nd.array([0.2822230782496945]),
        "variance": mx.nd.array([0.07964986589673317]),
    },
    "OneInflatedBeta": {
        "mean": mx.nd.array([0.45806451612903226]),
        "stddev": mx.nd.array([0.44137416804715]),
        "variance": mx.nd.array([0.19481115621931383]),
    },
    "ZeroAndOneInflatedBeta": {
        "mean": mx.nd.array([0.3903225806451613]),
        "stddev": mx.nd.array([0.45545503304667967]),
        "variance": mx.nd.array([0.20743928712755205]),
    },
    "EmpiricalDistribution": {
        "mean": mx.nd.array([9.0, 55.0]),
        "stddev": mx.nd.array([5.7445626, 28.722813]),
        "variance": mx.nd.array([33.0, 825.0]),
    },
}

test_cases_quantile = [
    (
        Gaussian,
        {
            "mu": mx.nd.array([0.0]),
            "sigma": mx.nd.array([1.0]),
        },
    ),
    (
        GenPareto,
        {
            "xi": mx.nd.array([1 / 3.0]),
            "beta": mx.nd.array([1.0]),
        },
    ),
    (
        EmpiricalDistribution,
        {
            "samples": mx.nd.stack(
                *[
                    mx.nd.arange(start=0, stop=20, step=2),
                    mx.nd.arange(start=100, stop=0, step=-10),
                ]
            ).transpose(),
            "event_dim": 1,
        },
    ),
]

test_output_quantile = {
    "Gaussian": {
        "x": mx.nd.array([3.0902362]),
        "cdf": mx.nd.array([0.999]),
        "level": mx.nd.array([0.999]),
        "quantile": mx.nd.array([[3.0902362]]),
    },
    "GenPareto": {
        "x": mx.nd.array([26.99999999999998]),
        "cdf": mx.nd.array([0.999]),
        "level": mx.nd.array([0.999]),
        "quantile": mx.nd.array([[26.99999999999998]]),
    },
    "EmpiricalDistribution": {
        # Actual samples are (transpose of):
        # [[  0.   2.   4.   6.   8.  10.  12.  14.  16.  18.]
        #  [100.  90.  80.  70.  60.  50.  40.  30.  20.  10.]]
        "x": mx.nd.array([2, 80]),
        "cdf": mx.nd.array([0.2, 0.8]),
        "level": mx.nd.array([0.1]),
        "quantile": mx.nd.array([[0, 10]]),
    },
}

# TODO: implement stddev methods for MultivariateGaussian and LowrankMultivariateGaussian
DISTRIBUTIONS = [
    Gaussian,
    Laplace,
    StudentT,
    Gamma,
    NegativeBinomial,
    Uniform,
    Binned,
    Poisson,
    ZeroInflatedBeta,
    OneInflatedBeta,
    ZeroAndOneInflatedBeta,
]


serialize_fn_list = [lambda x: x, lambda x: load_json(dump_json(x))]


@pytest.mark.parametrize("distr_class, params", test_cases)
@pytest.mark.parametrize("serialize_fn", serialize_fn_list)
def test_means(distr_class, params, serialize_fn) -> None:
    distr = distr_class(**params)
    distr = serialize_fn(distr)
    means = distr.mean
    distr_name = distr.__class__.__name__
    assert means.shape == test_output[distr_name]["mean"].shape
    # asnumpy()  needed to b/c means is all pointers to values
    assert np.allclose(
        means.asnumpy(), test_output[distr_name]["mean"].asnumpy(), atol=1e-11
    )


@pytest.mark.parametrize("distr_class, params", test_cases)
@pytest.mark.parametrize("serialize_fn", serialize_fn_list)
def test_stdevs(distr_class, params, serialize_fn) -> None:
    distr = distr_class(**params)
    distr = serialize_fn(distr)
    stddevs = distr.stddev
    distr_name = distr.__class__.__name__
    assert stddevs.shape == test_output[distr_name]["stddev"].shape
    assert np.allclose(
        stddevs.asnumpy(),
        test_output[distr_name]["stddev"].asnumpy(),
        atol=1e-11,
    )


@pytest.mark.parametrize("distr_class, params", test_cases)
@pytest.mark.parametrize("serialize_fn", serialize_fn_list)
def test_variances(distr_class, params, serialize_fn) -> None:
    distr = distr_class(**params)
    distr = serialize_fn(distr)
    variances = distr.variance
    distr_name = distr.__class__.__name__
    assert variances.shape == test_output[distr_name]["variance"].shape
    assert np.allclose(
        variances.asnumpy(),
        test_output[distr_name]["variance"].asnumpy(),
        atol=1e-11,
    )


@pytest.mark.parametrize("distr_class, params", test_cases_quantile)
@pytest.mark.parametrize("serialize_fn", serialize_fn_list)
def test_quantile(distr_class, params, serialize_fn) -> None:
    distr = distr_class(**params)
    distr = serialize_fn(distr)
    distr_name = distr.__class__.__name__
    quantile = distr.quantile(test_output_quantile[distr_name]["level"])
    assert np.allclose(
        quantile.asnumpy(),
        test_output_quantile[distr_name]["quantile"].asnumpy(),
        atol=1e-11,
    )


@pytest.mark.parametrize("distr_class, params", test_cases_quantile)
@pytest.mark.parametrize("serialize_fn", serialize_fn_list)
def test_cdf(distr_class, params, serialize_fn) -> None:
    distr = distr_class(**params)
    distr = serialize_fn(distr)
    distr_name = distr.__class__.__name__
    cdf = distr.cdf(test_output_quantile[distr_name]["x"])
    assert np.allclose(
        cdf.asnumpy(),
        test_output_quantile[distr_name]["cdf"].asnumpy(),
        atol=1e-11,
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from gluonts.core import serde
from gluonts.dataset.repository import datasets
from gluonts.dataset.util import forecast_start, to_pandas
from gluonts.evaluation import Evaluator, backtest_metrics
from gluonts.model import QuantileForecast
from gluonts.ext.r_forecast import (
    RForecastPredictor,
    R_IS_INSTALLED,
    RPY2_IS_INSTALLED,
    SUPPORTED_UNIVARIATE_METHODS,
    UNIVARIATE_POINT_FORECAST_METHODS,
)


# conditionally skip these tests if `R` and `rpy2` are not installed
if not R_IS_INSTALLED or not RPY2_IS_INSTALLED:
    skip_message = "Skipping test because `R` and `rpy2` are not installed!"
    pytest.skip(skip_message, allow_module_level=True)


TOLERANCE = 1e-6


@pytest.mark.parametrize("method_name", SUPPORTED_UNIVARIATE_METHODS)
def test_forecasts(method_name):
    if method_name == "mlp":
        # https://stackoverflow.com/questions/56254321/error-in-ifncol-matrix-rep-argument-is-of-length-zero
        # https://cran.r-project.org/web/packages/neuralnet/index.html
        #   published before the bug fix: https://github.com/bips-hb/neuralnet/pull/21
        # The issue is still open on nnfor package: https://github.com/trnnick/nnfor/issues/8
        # TODO: look for a workaround.
        pytest.xfail(
            "MLP currently does not work because "
            "the `neuralnet` package is not yet updated with a known bug fix in ` bips-hb/neuralnet`"
        )

    dataset = datasets.get_dataset("constant")

    freq = dataset.metadata.freq
    prediction_length = dataset.metadata.prediction_length

    predictor = RForecastPredictor(
        freq=freq,
        prediction_length=prediction_length,
        method_name=method_name,
    )
    predictions = list(predictor.predict(dataset.train))

    assert all(
        isinstance(prediction, QuantileForecast) for prediction in predictions
    )

    assert all(prediction.freq == freq for prediction in predictions)

    assert all(
        prediction.prediction_length == prediction_length
        for prediction in predictions
    )

    assert all(
        prediction.start_date == forecast_start(data)
        for data, prediction in zip(dataset.train, predictions)
    )

    evaluator = Evaluator(
        allow_nan_forecast=method_name in UNIVARIATE_POINT_FORECAST_METHODS
    )
    agg_metrics, item_metrics = backtest_metrics(
        test_dataset=dataset.test,
        predictor=predictor,
        evaluator=evaluator,
    )
    assert agg_metrics["mean_wQuantileLoss"] < TOLERANCE
    assert agg_metrics["NRMSE"] < TOLERANCE
    assert agg_metrics["RMSE"] < TOLERANCE

    predictor = RForecastPredictor(
        freq=freq,
        prediction_length=prediction_length,
        method_name=method_name,
        trunc_length=prediction_length,
    )
    predictions = list(predictor.predict(dataset.train))

    assert all(
        prediction.start_date == to_pandas(data).index[-1] + 1
        for data, prediction in zip(dataset.train, predictions)
    )


def test_r_predictor_serialization():
    predictor = RForecastPredictor(freq="1D", prediction_length=3)
    assert predictor == serde.decode(serde.encode(predictor))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

from gluonts.core.component import validated
from gluonts.mx import Tensor

from .representation import Representation


class DimExpansion(Representation):
    """
    A class representing a dimension expansion operation along a specified
    axis.

    Parameters
    ----------
    axis
        Axis on which to expand the tensor.
        (default: -1)
    """

    @validated()
    def __init__(self, axis: int = -1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.axis = axis

    def hybrid_forward(
        self,
        F,
        data: Tensor,
        observed_indicator: Tensor,
        scale: Optional[Tensor],
        rep_params: List[Tensor],
        **kwargs,
    ) -> Tuple[Tensor, Tensor, List[Tensor]]:
        data = F.expand_dims(data, axis=self.axis)
        return data, scale, rep_params



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from gluonts.dataset.repository.datasets import (
    get_dataset,
    get_download_path,
    dataset_recipes,
    default_dataset_path,
    dataset_names,
)

__all__ = [
    "get_dataset",
    "get_download_path",
    "dataset_recipes",
    "default_dataset_path",
    "dataset_names",
]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from gluonts.mx import MQCNNEstimator, MQRNNEstimator
from gluonts.mx.distribution import GaussianOutput, ISQFOutput
from gluonts.testutil.dummy_datasets import make_dummy_datasets_with_features


@pytest.fixture()
def hyperparameters(dsinfo):
    return dict(
        ctx="cpu",
        epochs=1,
        learning_rate=1e-2,
        hybridize=True,
        context_length=dsinfo.prediction_length,
        num_forking=1,
        num_batches_per_epoch=1,
        quantiles=[0.1, 0.5, 0.9],
        use_symbol_block_predictor=True,
    )


@pytest.fixture(
    params=[MQCNNEstimator, MQRNNEstimator], ids=["mqcnn", "mqrnn"]
)
def Estimator(request):
    return request.param


@pytest.mark.parametrize("hybridize", [True, False])
@pytest.mark.parametrize(
    "quantiles, distr_output",
    [
        ([0.1, 0.5, 0.9], None),
        (None, GaussianOutput()),
        (None, ISQFOutput(1, [0.1, 0.9])),
        (None, ISQFOutput(5, [0.1, 0.3, 0.5, 0.7, 0.9])),
    ],
)
@pytest.mark.parametrize("is_iqf", [True, False])
def test_accuracy(
    Estimator,
    accuracy_test,
    hyperparameters,
    hybridize,
    quantiles,
    distr_output,
    is_iqf,
):
    hyperparameters.update(
        num_batches_per_epoch=100,
        hybridize=hybridize,
        quantiles=quantiles,
        distr_output=distr_output,
        is_iqf=is_iqf,
    )

    accuracy_test(
        Estimator, hyperparameters, accuracy=0.20 if quantiles else 0.70
    )


@pytest.mark.parametrize("use_past_feat_dynamic_real", [True, False])
@pytest.mark.parametrize("use_feat_dynamic_real", [True, False])
@pytest.mark.parametrize("add_time_feature", [True, False])
@pytest.mark.parametrize("add_age_feature", [True, False])
@pytest.mark.parametrize("enable_encoder_dynamic_feature", [True, False])
@pytest.mark.parametrize("enable_decoder_dynamic_feature", [True, False])
@pytest.mark.parametrize("hybridize", [True, False])
@pytest.mark.parametrize(
    "quantiles, distr_output",
    [
        ([0.5, 0.1], None),
        (None, GaussianOutput()),
        (None, ISQFOutput(1, [0.1, 0.9])),
        (None, ISQFOutput(5, [0.1, 0.3, 0.5, 0.7, 0.9])),
    ],
)
@pytest.mark.parametrize("is_iqf", [True, False])
def test_mqcnn_covariate_smoke_test(
    use_past_feat_dynamic_real,
    use_feat_dynamic_real,
    add_time_feature,
    add_age_feature,
    enable_encoder_dynamic_feature,
    enable_decoder_dynamic_feature,
    hybridize,
    quantiles,
    distr_output,
    is_iqf,
):
    hps = {
        "seed": 42,
        "freq": "Y",
        "context_length": 5,
        "prediction_length": 3,
        "quantiles": quantiles,
        "distr_output": distr_output,
        "epochs": 1,
        "num_batches_per_epoch": 1,
        "use_past_feat_dynamic_real": use_past_feat_dynamic_real,
        "use_feat_dynamic_real": use_feat_dynamic_real,
        "add_time_feature": add_time_feature,
        "add_age_feature": add_age_feature,
        "enable_encoder_dynamic_feature": enable_encoder_dynamic_feature,
        "enable_decoder_dynamic_feature": enable_decoder_dynamic_feature,
        "hybridize": hybridize,
        "is_iqf": is_iqf,
    }

    dataset_train, dataset_test = make_dummy_datasets_with_features(
        cardinality=[3, 10],
        num_feat_dynamic_real=2,
        num_past_feat_dynamic_real=4,
        freq=hps["freq"],
        prediction_length=hps["prediction_length"],
    )

    estimator = MQCNNEstimator.from_hyperparameters(**hps)

    predictor = estimator.train(dataset_train, num_workers=None)
    forecasts = list(predictor.predict(dataset_test))
    assert len(forecasts) == len(dataset_test)


@pytest.mark.parametrize("use_feat_static_cat", [True, False])
@pytest.mark.parametrize("cardinality", [[], [3, 10]])
def test_feat_static_cat_smoke_test(use_feat_static_cat, cardinality):
    hps = {
        "seed": 42,
        "freq": "D",
        "prediction_length": 3,
        "quantiles": [0.5, 0.1],
        "epochs": 1,
        "num_batches_per_epoch": 1,
        "use_feat_static_cat": use_feat_static_cat,
    }

    dataset_train, dataset_test = make_dummy_datasets_with_features(
        cardinality=cardinality,
        num_feat_dynamic_real=2,
        freq=hps["freq"],
        prediction_length=hps["prediction_length"],
    )
    estimator = MQCNNEstimator.from_inputs(dataset_train, **hps)

    predictor = estimator.train(dataset_train, num_workers=None)
    forecasts = list(predictor.predict(dataset_test))
    assert len(forecasts) == len(dataset_test)


# Test scaling and from inputs
@pytest.mark.parametrize("scaling", [True, False])
@pytest.mark.parametrize("scaling_decoder_dynamic_feature", [True, False])
def test_mqcnn_scaling_smoke_test(scaling, scaling_decoder_dynamic_feature):
    hps = {
        "seed": 42,
        "freq": "D",
        "prediction_length": 3,
        "quantiles": [0.5, 0.1],
        "epochs": 1,
        "num_batches_per_epoch": 1,
        "scaling": scaling,
        "scaling_decoder_dynamic_feature": scaling_decoder_dynamic_feature,
    }

    dataset_train, dataset_test = make_dummy_datasets_with_features(
        cardinality=[3, 10],
        num_feat_dynamic_real=2,
        freq=hps["freq"],
        prediction_length=hps["prediction_length"],
    )

    estimator = MQCNNEstimator.from_inputs(dataset_train, **hps)

    predictor = estimator.train(dataset_train, num_workers=None)
    forecasts = list(predictor.predict(dataset_test))
    assert len(forecasts) == len(dataset_test)


@pytest.mark.parametrize(
    "quantiles, inference_quantiles",
    [([0.5, 0.1, 0.9], [0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 0.99])],
)
def test_inference_quantile_prediction(quantiles, inference_quantiles):
    hps = {
        "seed": 42,
        "freq": "D",
        "prediction_length": 3,
        "quantiles": quantiles,
        "epochs": 1,
        "num_batches_per_epoch": 1,
        "is_iqf": True,
    }

    dataset_train, dataset_test = make_dummy_datasets_with_features(
        cardinality=[3, 10],
        num_feat_dynamic_real=2,
        freq=hps["freq"],
        prediction_length=hps["prediction_length"],
    )

    estimator = MQCNNEstimator.from_inputs(dataset_train, **hps)

    predictor = estimator.train(dataset_train, num_workers=None)
    forecasts = list(predictor.predict(dataset_test))
    assert len(forecasts) == len(dataset_test)
    item_id = 0
    for inference_quantile in inference_quantiles:
        assert (
            len(forecasts[item_id].quantile(inference_quantile))
            == hps["prediction_length"]
        )
    inference_quantile = sorted(inference_quantiles)
    previous_quantile_prediction = forecasts[item_id].quantile(
        inference_quantile[0]
    )
    for inference_quantile in inference_quantiles[1:]:
        assert all(
            previous_quantile_prediction[i] <= pred
            for (i, pred) in enumerate(
                forecasts[item_id].quantile(inference_quantile)
            )
        ), "quantile-crossing occurred"


@pytest.mark.parametrize("is_iqf", [True, False])
def test_is_iqf(is_iqf):
    hps = {
        "seed": 42,
        "freq": "D",
        "prediction_length": 3,
        "quantiles": [0.5, 0.1],
        "epochs": 1,
        "num_batches_per_epoch": 1,
        "is_iqf": is_iqf,
    }

    dataset_train, dataset_test = make_dummy_datasets_with_features(
        cardinality=[3, 10],
        num_feat_dynamic_real=2,
        freq=hps["freq"],
        prediction_length=hps["prediction_length"],
    )

    estimator = MQCNNEstimator.from_inputs(dataset_train, **hps)

    predictor = estimator.train(dataset_train, num_workers=None)
    forecasts = list(predictor.predict(dataset_test))
    assert len(forecasts) == len(dataset_test)


@pytest.mark.parametrize("is_iqf", [True, False])
def test_serialize(Estimator, serialize_test, hyperparameters, is_iqf):
    hyperparameters.update(is_iqf=is_iqf)
    serialize_test(Estimator, hyperparameters)


def test_backwards_compatibility():
    hps = {
        "freq": "D",
        "context_length": 5,
        "num_forking": 4,
        "prediction_length": 3,
        "quantiles": [0.5, 0.1],
        "epochs": 1,
        "num_batches_per_epoch": 1,
        "use_feat_dynamic_real": True,
        "use_past_feat_dynamic_real": True,
        "enable_encoder_dynamic_feature": True,
        "enable_decoder_dynamic_feature": True,
        "scaling": True,
        "scaling_decoder_dynamic_feature": True,
    }

    dataset_train, dataset_test = make_dummy_datasets_with_features(
        cardinality=[3, 10],
        num_feat_dynamic_real=2,
        num_past_feat_dynamic_real=4,
        freq=hps["freq"],
        prediction_length=hps["prediction_length"],
    )

    dataset_train = list(dataset_train)
    dataset_test = list(dataset_test)

    for dataset in dataset_train, dataset_test:
        for entry in dataset:
            entry["dynamic_feat"] = entry.pop("feat_dynamic_real")

    estimator = MQCNNEstimator.from_inputs(dataset_train, **hps)

    predictor = estimator.train(dataset_train, num_workers=None)
    forecasts = list(predictor.predict(dataset_test))
    assert len(forecasts) == len(dataset_test)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from mxnet.gluon import HybridBlock, nn

from gluonts.core.component import validated
from gluonts.mx.block.rnn import RNN


class RNNModel(HybridBlock):
    @validated()
    def __init__(
        self,
        mode,
        num_hidden,
        num_layers,
        num_output,
        bidirectional=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_output = num_output

        with self.name_scope():
            self.rnn = RNN(
                mode=mode,
                num_hidden=num_hidden,
                num_layers=num_layers,
                bidirectional=bidirectional,
            )

            self.decoder = nn.Dense(
                num_output, in_units=num_hidden, flatten=False
            )

    def hybrid_forward(self, F, inputs):
        return self.decoder(self.rnn(inputs))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
gluonts.core.serde
~~~~~~~~~~~~~~~~~~~~~

Simple Serialization/Deserialization framework that uses human-readable
json-encodings to represent data.

In some ways, it is an alternative to `pickle`, but instead of using a binary
encoding, `serde` outputs json-compatible text. This has two main advantages:
It is human-readable and the encoding works across different Python versions.

Values which can be directly represented in `json` are not changed, thus
`serde.encode` is virtually the same as `json.dumps` for values such as `1`,
`["a", "b", "c"]` or `{"foo": "bar"}`.

However, more complex values are encoded differently. For these, we create an
object which uses a special `__kind__` attribute, which can be one of::

    class Kind(str, Enum):
        Type = "type"
        Instance = "instance"
        Stateful = "stateful"


A type is just a path to a class or function, an instance is an object which
can be re-constructed by passing arguments to its constructor, and stateful
represents objects that are decoded by setting the `__dict__` attribute on
an otherwise empty instance of the type.

`serde.encode` uses `functools.singledispatch` to encode a given object. It is
implemented for a variety of existing types, such as named-tuples, paths, and
`pydantic.BaseModel`.

In addition, one can derive from `serde.Stateful` or `serde.Stateless` to opt
into one of the behaviours. The latter requires that the class supports
`__getnewargs_ex__` from the pickle protocol.

To encode custom values, one can use the `serde.encode.register` decorator.
For example, support for numpy's arrays are added by::

    @encode.register(np.ndarray)
    def encode_np_ndarray(v: np.ndarray) -> Any:
        return {
            "__kind__": Kind.Instance,
            "class": "numpy.array",
            "args": encode([v.tolist(), v.dtype]),
        }

There is no need to implement `decode` for a given type since the encoding
should contain the information on how the object can be constructed.

Similarly to `json`, `serde` does not support object-identity. This means that
if an object exists twice in an object graph, it is encoded at each of the
occurances. Consequently, ciruclar references do not work.

`dump_json` and `load_json` are simple helpers, which use `encode` and
`decode` internally.
"""

from . import flat
from ._base import Stateful, Stateless, decode, encode
from ._dataclass import dataclass, EVENTUAL, Eventual, OrElse
from ._json import dump_json, load_json


# TODO: remove
# These are needed because we implement `encode` for numpy and pandas types in
# submodules.
from .np import *  # noqa
from .pd import *  # noqa


__all__ = [
    "flat",
    "encode",
    "decode",
    "dump_json",
    "load_json",
    "Stateful",
    "Stateless",
    "dataclass",
    "EVENTUAL",
    "Eventual",
    "OrElse",
]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pandas as pd
import numpy as np


def assert_recursively_equal(obj_a, obj_b, equal_nan=True):
    """
    Asserts that two objects are equal, recursively.

    This is based on :func:`assert_recursively_close`, and accepts the
    same arguments, except that tolerances are set to zero.

    Parameters:
    -----------
    obj_a
    obj_b
        Objects to compare.
    equal_nan
        Indicates whether or not numpy.nan values should be considered equal.
    """
    _assert_recursively_close(
        obj_a, obj_b, location="", rtol=0, atol=0, equal_nan=equal_nan
    )


def assert_recursively_close(
    obj_a, obj_b, rtol=1e-05, atol=1e-08, equal_nan=True
):
    """
    Asserts that two objects are "close" to each other, recursively.

    Strings or ints are close iff they are equal; floats or numpy arrays
    are defined close according to the numpy.isclose and numpy.allclose
    functions, respectively. Lists are close if all of their items are close.
    Dicts are close if they have the same keys, and elements corresponding
    to the same key are close.

    Parameters:
    -----------
    obj_a
    obj_b
        Objects to compare.
    rtol
    atol
        Relative and absolute tolerance for float comparison; see docs for
        numpy.isclose.
    equal_nan
        Indicates whether or not numpy.nan values should be considered equal.
    """
    _assert_recursively_close(
        obj_a, obj_b, location="", rtol=rtol, atol=atol, equal_nan=equal_nan
    )


def _assert_recursively_close(obj_a, obj_b, location, *args, **kwargs):
    assert type(obj_a) == type(
        obj_b
    ), f"types don't match (at {location}) {type(obj_a)} != {type(obj_b)}"
    if isinstance(obj_a, (str, int)):
        assert obj_a == obj_b
    elif isinstance(obj_a, float):
        assert np.isclose(obj_a, obj_b, *args, **kwargs)
    elif isinstance(obj_a, list):
        assert len(obj_a) == len(obj_b), f"lengths don't match (at {location})"
        for i, (element_a, element_b) in enumerate(zip(obj_a, obj_b)):
            _assert_recursively_close(
                element_a,
                element_b,
                location=f"{location}.{i}",
                *args,
                **kwargs,
            )
    elif isinstance(obj_a, dict):
        assert (
            obj_a.keys() == obj_b.keys()
        ), f"keys don't match (at {location})"
        for k in obj_a:
            _assert_recursively_close(
                obj_a[k], obj_b[k], location=f"{location}.{k}", *args, **kwargs
            )
    elif isinstance(obj_a, np.ndarray):
        assert (
            obj_a.dtype == obj_b.dtype
        ), f"numpy arrays have different dtype (at {location})"
        assert np.allclose(
            obj_a, obj_b, *args, **kwargs
        ), f"numpy arrays are not close enough (at {location})"
    elif isinstance(obj_a, pd.Period):
        assert obj_a == obj_b
    elif obj_a is None:
        assert obj_b is None
    else:
        raise TypeError(f"unsupported type {type(obj_a)}")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np

from gluonts.dataset.common import ListDataset
from gluonts.evaluation import backtest_metrics
from gluonts.model.predictor import Localizer, ParallelizedPredictor
from gluonts.model.trivial.identity import IdentityPredictor
from gluonts.model.trivial.mean import MeanEstimator


def test_parallelized_predictor():
    dataset = ListDataset(
        data_iter=[
            {"start": "2012-01-01", "target": (np.zeros(20) + i).tolist()}
            for i in range(300)
        ],
        freq="1H",
    )

    base_predictor = IdentityPredictor(prediction_length=10, num_samples=100)

    predictor = ParallelizedPredictor(
        base_predictor=base_predictor, num_workers=10, chunk_size=2
    )

    predictions = list(base_predictor.predict(dataset))
    parallel_predictions = list(predictor.predict(dataset))

    assert len(predictions) == len(parallel_predictions)

    for p, pp in zip(predictions, parallel_predictions):
        assert np.all(p.samples == pp.samples)
        assert np.all(p.index == pp.index)


def test_localizer():
    dataset = ListDataset(
        data_iter=[
            {
                "start": "2012-01-01",
                "target": (np.zeros(20) + i * 0.1 + 0.01),
                "id": f"{i}",
            }
            for i in range(3)
        ],
        freq="1H",
    )

    estimator = MeanEstimator(prediction_length=10, num_samples=50)

    local_pred = Localizer(estimator=estimator)
    agg_metrics, _ = backtest_metrics(
        test_dataset=dataset, predictor=local_pred
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from gluonts.dataset.artificial import constant_dataset
from gluonts.dataset.common import TrainDatasets
from gluonts.dataset.multivariate_grouper import MultivariateGrouper
from gluonts.evaluation import backtest_metrics, MultivariateEvaluator
from gluonts.mx import DeepVAREstimator
from gluonts.mx.distribution import (
    LowrankMultivariateGaussianOutput,
    MultivariateGaussianOutput,
)
from gluonts.mx.trainer import Trainer


def load_multivariate_constant_dataset():
    dataset_info, train_ds, test_ds = constant_dataset()
    grouper_train = MultivariateGrouper(max_target_dim=10)
    grouper_test = MultivariateGrouper(num_test_dates=1, max_target_dim=10)
    metadata = dataset_info.metadata
    metadata.prediction_length = dataset_info.prediction_length
    return TrainDatasets(
        metadata=dataset_info.metadata,
        train=grouper_train(train_ds),
        test=grouper_test(test_ds),
    )


dataset = load_multivariate_constant_dataset()
target_dim = int(dataset.metadata.feat_static_cat[0].cardinality)
metadata = dataset.metadata
estimator = DeepVAREstimator


@pytest.mark.parametrize(
    "distr_output, num_batches_per_epoch, Estimator, hybridize, "
    "use_marginal_transformation",
    [
        (
            LowrankMultivariateGaussianOutput(dim=target_dim, rank=2),
            10,
            estimator,
            True,
            True,
        ),
        (
            LowrankMultivariateGaussianOutput(dim=target_dim, rank=2),
            10,
            estimator,
            False,
            False,
        ),
        (
            LowrankMultivariateGaussianOutput(dim=target_dim, rank=2),
            10,
            estimator,
            True,
            False,
        ),
        (None, 10, estimator, True, True),
        (
            MultivariateGaussianOutput(dim=target_dim),
            10,
            estimator,
            False,
            True,
        ),
        (
            MultivariateGaussianOutput(dim=target_dim),
            10,
            estimator,
            True,
            True,
        ),
    ],
)
@pytest.mark.parametrize("use_feat_dynamic_real", [True, False])
def test_deepvar(
    distr_output,
    num_batches_per_epoch,
    Estimator,
    hybridize,
    use_marginal_transformation,
    use_feat_dynamic_real,
):
    estimator = Estimator(
        num_cells=20,
        num_layers=1,
        pick_incomplete=True,
        target_dim=target_dim,
        prediction_length=metadata.prediction_length,
        # target_dim=target_dim,
        freq=metadata.freq,
        distr_output=distr_output,
        scaling=False,
        use_marginal_transformation=use_marginal_transformation,
        use_feat_dynamic_real=use_feat_dynamic_real,
        trainer=Trainer(
            epochs=1,
            learning_rate=1e-10,
            num_batches_per_epoch=num_batches_per_epoch,
            hybridize=hybridize,
        ),
    )

    predictor = estimator.train(training_data=dataset.train)

    agg_metrics, _ = backtest_metrics(
        test_dataset=dataset.test,
        predictor=predictor,
        evaluator=MultivariateEvaluator(
            quantiles=(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9)
        ),
    )

    assert agg_metrics["ND"] < 1.5



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from gluonts.mx.model.n_beats import NBEATSEnsembleEstimator, NBEATSEstimator


@pytest.fixture()
def args(dsinfo):
    common_hp = dict(
        ctx="cpu",
        epochs=1,
        hybridize=True,
        num_batches_per_epoch=1,
    )
    return {
        "generic": {
            "estimator": NBEATSEstimator,
            "args": {"num_stacks": 2},
            **common_hp,
        },
        "interpretable": {
            "estimator": NBEATSEstimator,
            "args": {
                "num_stacks": 2,
                "num_blocks": [3],
                "widths": [256, 2048],
                "sharing": [True],
                "stack_types": ["T", "S"],
                **common_hp,
            },
        },
        "ensemble": {
            "estimator": NBEATSEnsembleEstimator,
            "args": {
                "meta_context_length": [2 * dsinfo.prediction_length],
                "meta_loss_function": ["MAPE"],
                "meta_bagging_size": 2,
                "num_stacks": 3,
                **common_hp,
            },
        },
    }


@pytest.fixture(params=["generic", "interpretable", "ensemble"])
def name(request):
    return request.param


@pytest.fixture()
def estimator_config(args, name):
    config = args[name]
    return name, config["estimator"], config["args"]


# tests are too slow
@pytest.mark.skip("Skipping slow test.")
@pytest.mark.parametrize("hybridize", [True, False])
def test_accuracy(accuracy_test, estimator_config, hybridize):
    Estimator, hyperparameters = estimator_config[1:]
    hyperparameters.update(num_batches_per_epoch=200, hybridize=hybridize)

    accuracy_test(Estimator, hyperparameters, accuracy=0.3)


def test_serialize(serialize_test, estimator_config):
    if estimator_config[0] == "generic":
        pytest.skip("Too slow.")
    serialize_test(*estimator_config[1:])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Any


def fqname_for(obj: Any) -> str:
    """
    Returns the fully qualified name of ``obj``.

    Parameters
    ----------
    obj
        The class we are interested in.

    Returns
    -------
    str
        The fully qualified name of ``obj``.
    """

    if "<locals>" in obj.__qualname__:
        raise RuntimeError(
            "Can't get fully qualified name of locally defined object. "
            f"{obj.__qualname__}"
        )

    return f"{obj.__module__}.{obj.__qualname__}"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Optional, Tuple, Type

import mxnet as mx
import numpy as np

from gluonts.core.component import validated
from gluonts.mx import Tensor

from .distribution import Distribution, _sample_multiple, getF
from .distribution_output import DistributionOutput


class Deterministic(Distribution):
    r"""
    Deterministic/Degenerate distribution.
    Parameters
    ----------
    value
        Tensor containing the values, of shape `(*batch_shape, *event_shape)`.
    F
    """

    is_reparameterizable = True

    @validated()
    def __init__(self, value: Tensor) -> None:
        self.value = value

    @property
    def F(self):
        return getF(self.value)

    @property
    def batch_shape(self) -> Tuple:
        return self.value.shape

    @property
    def event_shape(self) -> Tuple:
        return ()

    @property
    def event_dim(self) -> int:
        return 0

    def log_prob(self, x: Tensor) -> Tensor:
        F = self.F
        value = self.value
        is_both_nan = F.broadcast_logical_and(x != x, value != value)
        is_equal_or_both_nan = F.broadcast_logical_or(
            (x == value), is_both_nan
        )
        return F.log(is_equal_or_both_nan)

    @property
    def mean(self) -> Tensor:
        return self.value

    @property
    def stddev(self) -> Tensor:
        return self.value.zeros_like()

    def cdf(self, x):
        F = self.F
        value = self.value
        is_both_nan = F.broadcast_logical_and(
            F.contrib.isnan(x), F.contrib.isnan(value)
        )
        is_greater_equal_or_both_nan = F.broadcast_logical_or(
            (x >= value), is_both_nan
        )
        return is_greater_equal_or_both_nan

    def sample(
        self, num_samples: Optional[int] = None, dtype=np.float32
    ) -> Tensor:
        return _sample_multiple(
            lambda value: value, value=self.value, num_samples=num_samples
        ).astype(dtype=dtype)

    def quantile(self, level: Tensor) -> Tensor:
        F = self.F
        # we consider level to be an independent axis and so expand it
        # to shape (num_levels, 1, 1, ...)

        for _ in range(self.all_dim):
            level = level.expand_dims(axis=-1)

        quantiles = F.broadcast_mul(self.value, level.ones_like())
        level = F.broadcast_mul(quantiles.ones_like(), level)

        minus_inf = -quantiles.ones_like() / 0.0
        quantiles = F.where(
            F.broadcast_logical_or(level != 0, F.contrib.isnan(quantiles)),
            quantiles,
            minus_inf,
        )

        nans = level.zeros_like() / 0.0
        quantiles = F.where(level != level, nans, quantiles)

        return quantiles

    @property
    def args(self) -> List:
        return [self.value]


class DeterministicArgProj(mx.gluon.HybridBlock):
    def __init__(
        self,
        value: float,
        args_dim: Dict[str, int],
        dtype: Type = np.float32,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.value = value
        self.args_dim = args_dim
        self.dtype = dtype

    def hybrid_forward(self, F, x: Tensor) -> Tuple[Tensor]:
        return (self.value * F.ones_like(x.sum(axis=-1)),)


class DeterministicOutput(DistributionOutput):
    args_dim: Dict[str, int] = {"value": 1}
    distr_cls: type = Deterministic

    @validated()
    def __init__(self, value: float):
        super().__init__()
        self.value = value

    def get_args_proj(
        self, prefix: Optional[str] = None
    ) -> mx.gluon.HybridBlock:
        return DeterministicArgProj(
            value=self.value, args_dim=self.args_dim, dtype=self.dtype
        )

    @property
    def event_shape(self) -> Tuple:
        return ()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import numpy as np
import pandas as pd
from toolz.itertoolz import first

import pytest
from gluonts.dataset.hierarchical import HierarchicalTimeSeries


PERIODS = 24
FREQ = "H"


def random_ts(num_ts: int, periods: int, freq: str):
    index = pd.period_range(start="22-03-2020", periods=periods, freq=freq)

    return pd.concat(
        [
            pd.Series(data=np.random.random(size=len(index)), index=index)
            for _ in range(num_ts)
        ],
        axis=1,
    )


def test_three_level_hierarchy():
    # Simple three-level hierarchy containing 4 leaf nodes.
    S = np.vstack(([[1, 1, 1, 1], [1, 1, 0, 0], [0, 0, 1, 1]], np.eye(4)))

    num_ts, num_bottom_ts = S.shape
    ts_at_bottom_level = random_ts(
        num_ts=num_bottom_ts,
        periods=PERIODS,
        freq=FREQ,
    )

    hts = HierarchicalTimeSeries(ts_at_bottom_level=ts_at_bottom_level, S=S)

    ts_at_all_levels = hts.ts_at_all_levels
    assert (ts_at_all_levels.index == ts_at_bottom_level.index).all(), (
        "The index of dataframe `ts_at_all_levels` does not match "
        "with that of  `ts_at_bottom_level`:\n"
        f"Index of `ts_at_bottom_level`: {ts_at_bottom_level.index}, \n "
        f"Index of `ts_at_all_levels`: {ts_at_all_levels.index}."
    )

    assert ts_at_all_levels.shape == (PERIODS, num_ts), (
        "Hierarchical time series do not have the right shape. "
        f"Expected: {(PERIODS, num_ts)}, "
        f"Obtained: {ts_at_bottom_level.shape}!"
    )

    root_level = hts.ts_at_all_levels.iloc[:, 0]
    root_level_expected = ts_at_bottom_level.sum(axis=1)
    np.testing.assert_array_almost_equal(
        root_level.values,
        root_level_expected.values,
        err_msg="Values of the time series at the root"
        "level are not correctly computed.",
    )

    level_1 = hts.ts_at_all_levels.iloc[:, 1:3]
    level_1_expected = pd.concat(
        [
            ts_at_bottom_level.iloc[:, :2].sum(axis=1),
            ts_at_bottom_level.iloc[:, 2:].sum(axis=1),
        ],
        axis=1,
    )
    np.testing.assert_array_almost_equal(
        level_1.values,
        level_1_expected.values,
        err_msg="Values of the time series at the first"
        "aggregated level (after the root) are not "
        "correctly computed.",
    )

    leaf_level = hts.ts_at_all_levels.iloc[:, 3:]
    np.testing.assert_array_almost_equal(
        ts_at_bottom_level.values,
        leaf_level.values,
        err_msg="Values of the time series at the bottom "
        "level do not agree with the given inputs.",
    )


def get_random_hts(S: np.ndarray, periods: int, freq: str):
    num_ts, num_bottom_ts = S.shape
    ts_at_bottom_level = random_ts(
        num_ts=num_bottom_ts,
        periods=periods,
        freq=freq,
    )

    hts = HierarchicalTimeSeries(ts_at_bottom_level=ts_at_bottom_level, S=S)
    return hts


@pytest.mark.parametrize("mode", ["train", "inference", "fail"])
def test_hts_to_dataset(mode: str):
    S = np.vstack(([[1, 1, 1, 1], [1, 1, 0, 0], [0, 0, 1, 1]], np.eye(4)))
    hts = get_random_hts(S=S, periods=PERIODS, freq=FREQ)

    num_bottom_ts = S.shape[1]
    num_features = 10
    num_future_time_steps = {
        "train": 0,
        "inference": PERIODS // 2,
        "fail": PERIODS // 2,
    }[mode]

    features_df = random_ts(
        num_ts=num_features,
        periods=PERIODS + num_future_time_steps,
        freq=FREQ,
    )

    if mode == "fail":
        # Create a misalignment with the index of target time series.
        features_df.index = features_df.index.shift(periods=-1)

        with pytest.raises(Exception):
            ds = hts.to_dataset(feat_dynamic_real=features_df)
    else:
        ds = hts.to_dataset(feat_dynamic_real=features_df)
        entry = first(ds)

        assert entry["start"] == features_df.index[0]

        if mode == "train":
            entry["target"].shape == (num_bottom_ts, PERIODS)
            entry["feat_dynamic_real"].shape == (num_features, PERIODS)
        else:
            entry["target"].shape == (
                num_bottom_ts,
                PERIODS + num_future_time_steps,
            )
            entry["feat_dynamic_real"].shape == (
                num_features,
                PERIODS + num_future_time_steps,
            )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

import numpy as np

import gluonts.zebras as zb


def test_batch_time_series():
    ts0 = zb.time_series(
        [1, 2, 3], start="2020", freq="D", metadata={"item_id": 0}, name="x"
    )
    ts1 = zb.time_series([4, 5, 6])

    batched = zb.batch([ts0, ts1])

    assert len(batched) == 3
    assert batched.batch_size == 2

    ts0x, ts1x = batched.items()

    np.testing.assert_array_equal(ts0, ts0x)
    assert ts0.metadata == ts0x.metadata
    assert ts0.tdim == ts0x.tdim
    assert ts0.name == ts0x.name
    assert ts0._pad == ts0x._pad

    np.testing.assert_array_equal(ts1, ts1x)
    assert ts1.metadata == ts1x.metadata
    assert ts1.tdim == ts1x.tdim
    assert ts1.name == ts1x.name
    assert ts1._pad == ts1x._pad


def test_batch_time_frame():
    tf0 = zb.time_frame(
        {"x": [1, 2, 3]},
        start="2020",
        freq="D",
        metadata={"item_id": 0},
    )
    tf1 = zb.time_frame({"x": [4, 5, 6]})

    batched = zb.batch([tf0, tf1])

    assert len(batched) == 3
    assert batched.batch_size == 2

    tf0x, tf1x = batched.items()


def test_batch_split_frame():
    sf0 = zb.time_frame(
        {"x": [1, 2, 3]},
        start="2020",
        freq="D",
        metadata={"item_id": 0},
    ).split(1, past_length=4, future_length=3)
    sf1 = zb.time_frame({"x": [4, 5, 6]}).split(
        1, past_length=4, future_length=3
    )

    batched = zb.batch([sf0, sf1])

    assert len(batched) == 7
    assert batched.batch_size == 2

    sf0x, sf1x = batched.items()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations

import logging
from dataclasses import dataclass, field, InitVar
from typing import Any, Iterable, Optional, Type, Union, cast

import numpy as np
import pandas as pd
from pandas.core.indexes.datetimelike import DatetimeIndexOpsMixin
from toolz import first

from gluonts.maybe import Maybe
from gluonts.dataset.common import DataEntry
from gluonts.itertools import Map, StarMap, SizedIterable

logger = logging.getLogger(__name__)


@dataclass
class PandasDataset:
    """
    A dataset type based on ``pandas.DataFrame``.

    This class is constructed with a collection of ``pandas.DataFrame``
    objects where each ``DataFrame`` is representing one time series.
    Both ``target`` and ``timestamp`` columns are essential. Dynamic features
    of a series can be specified with together with the series' ``DataFrame``,
    while static features can be specified in a separate ``DataFrame`` object
    via the ``static_features`` argument.

    Parameters
    ----------
    dataframes
        Single ``pd.DataFrame``/``pd.Series`` or a collection as list or dict
        containing at least ``timestamp`` and ``target`` values.
        If a dict is provided, the key will be the associated ``item_id``.
    target
        Name of the column that contains the ``target`` time series.
        For multivariate targets, a list of column names should be provided.
    timestamp
        Name of the column that contains the timestamp information.
    freq
        Frequency of observations in the time series. Must be a valid pandas
        frequency.
    feat_dynamic_real
        List of column names that contain dynamic real features.
    past_feat_dynamic_real
        List of column names that contain dynamic real features only available
        in the past.
    static_features
        ``pd.DataFrame`` containing static features for the series. The index
        should contain the key of the series in the ``dataframes`` argument.
    future_length
        For target and past dynamic features last ``future_length``
        elements are removed when iterating over the data set.
    unchecked
        Whether consistency checks on indexes should be skipped.
        (Default: ``False``)
    assume_sorted
        Whether to assume that indexes are sorted by time, and skip sorting.
        (Default: ``False``)
    """

    dataframes: InitVar[
        Union[
            pd.DataFrame,
            pd.Series,
            Iterable[pd.DataFrame],
            Iterable[pd.Series],
            Iterable[tuple[Any, pd.DataFrame]],
            Iterable[tuple[Any, pd.Series]],
            dict[str, pd.DataFrame],
            dict[str, pd.Series],
        ]
    ]
    target: Union[str, list[str]] = "target"
    feat_dynamic_real: Optional[list[str]] = None
    past_feat_dynamic_real: Optional[list[str]] = None
    timestamp: Optional[str] = None
    freq: Optional[str] = None
    static_features: InitVar[Optional[pd.DataFrame]] = None
    future_length: int = 0
    unchecked: bool = False
    assume_sorted: bool = False
    dtype: Type = np.float32
    _data_entries: SizedIterable = field(init=False)
    _static_reals: pd.DataFrame = field(init=False)
    _static_cats: pd.DataFrame = field(init=False)

    def __post_init__(self, dataframes, static_features):
        if isinstance(dataframes, dict):
            pairs = dataframes.items()
        elif isinstance(dataframes, (pd.Series, pd.DataFrame)):
            pairs = [(None, dataframes)]
        else:
            assert isinstance(dataframes, SizedIterable)
            pairs = Map(pair_with_item_id, dataframes)

        self._data_entries = StarMap(self._pair_to_dataentry, pairs)

        if self.freq is None:
            assert (
                self.timestamp is None
            ), "You need to provide `freq` along with `timestamp`"

            self.freq = infer_freq(first(pairs)[1].index)

        static_features = Maybe(static_features).unwrap_or_else(pd.DataFrame)

        object_columns = static_features.select_dtypes(
            "object"
        ).columns.tolist()
        if object_columns:
            logger.warning(
                f"Columns {object_columns} in static_features "
                f"have 'object' as data type and will be ignored; "
                f"consider setting this to 'category' using pd.DataFrame.astype, "
                f"if you wish to use them as categorical columns."
            )

        self._static_reals = (
            static_features.select_dtypes("number").astype(self.dtype).T
        )
        self._static_cats = (
            static_features.select_dtypes("category")
            .apply(lambda col: col.cat.codes)
            .astype(self.dtype)
            .T
        )

    @property
    def num_feat_static_cat(self) -> int:
        return len(self._static_cats)

    @property
    def num_feat_static_real(self) -> int:
        return len(self._static_reals)

    @property
    def num_feat_dynamic_real(self) -> int:
        return Maybe(self.feat_dynamic_real).map_or(len, 0)

    @property
    def num_past_feat_dynamic_real(self) -> int:
        return Maybe(self.past_feat_dynamic_real).map_or(len, 0)

    @property
    def static_cardinalities(self):
        return self._static_cats.max(axis=1).values + 1

    def _pair_to_dataentry(self, item_id, df) -> DataEntry:
        if isinstance(df, pd.Series):
            df = df.to_frame(name=self.target)

        if self.timestamp:
            df.index = pd.DatetimeIndex(df[self.timestamp]).to_period(
                freq=self.freq
            )

        if not isinstance(df.index, pd.PeriodIndex):
            df = df.to_period(freq=self.freq)

        if not self.assume_sorted:
            df.sort_index(inplace=True)

        if not self.unchecked:
            assert is_uniform(df.index), (
                "Dataframe index is not uniformly spaced. "
                "If your dataframe contains data from multiple series in the "
                'same column ("long" format), consider constructing the '
                "dataset with `PandasDataset.from_long_dataframe` instead."
            )

        entry = {
            "start": df.index[0],
        }

        target = df[self.target].values
        target = target[: len(target) - self.future_length]
        entry["target"] = target.T

        if item_id is not None:
            entry["item_id"] = item_id

        if self.num_feat_static_cat > 0:
            entry["feat_static_cat"] = self._static_cats[item_id].values

        if self.num_feat_static_real > 0:
            entry["feat_static_real"] = self._static_reals[item_id].values

        if self.num_feat_dynamic_real > 0:
            entry["feat_dynamic_real"] = df[self.feat_dynamic_real].values.T

        if self.num_past_feat_dynamic_real > 0:
            past_feat_dynamic_real = df[self.past_feat_dynamic_real].values
            past_feat_dynamic_real = past_feat_dynamic_real[
                : len(past_feat_dynamic_real) - self.future_length
            ]
            entry["past_feat_dynamic_real"] = past_feat_dynamic_real.T

        return entry

    def __iter__(self):
        yield from self._data_entries
        self.unchecked = True

    def __len__(self) -> int:
        return len(self._data_entries)

    def __str__(self) -> str:
        info = ", ".join(
            [
                f"size={len(self)}",
                f"freq={self.freq}",
                f"num_feat_dynamic_real={self.num_feat_dynamic_real}",
                f"num_past_feat_dynamic_real={self.num_past_feat_dynamic_real}",
                f"num_feat_static_real={self.num_feat_static_real}",
                f"num_feat_static_cat={self.num_feat_static_cat}",
                f"static_cardinalities={self.static_cardinalities}",
            ]
        )
        return f"PandasDataset<{info}>"

    @classmethod
    def from_long_dataframe(
        cls,
        dataframe: pd.DataFrame,
        item_id: str,
        timestamp: Optional[str] = None,
        static_feature_columns: Optional[list[str]] = None,
        static_features: pd.DataFrame = pd.DataFrame(),
        **kwargs,
    ) -> "PandasDataset":
        """
        Construct ``PandasDataset`` out of a long data frame.

        A long dataframe contains time series data (both the target series and
        covariates) about multiple items at once. An ``item_id`` column is used
        to distinguish the items and ``group_by`` accordingly.

        Static features can be included in the long data frame as well (with
        constant value), or be given as a separate data frame indexed by the
        ``item_id`` values.

        Note: on large datasets, this constructor can take some time to complete
        since it does some indexing and groupby operations on the data, and caches
        the result.

        Parameters
        ----------
        dataframe
            pandas.DataFrame containing at least ``timestamp``, ``target`` and
            ``item_id`` columns.
        item_id
            Name of the column that, when grouped by, gives the different time
            series.
        static_feature_columns
            Columns in ``dataframe`` containing static features.
        static_features
            Dedicated ``DataFrame`` for static features. If both ``static_features``
            and ``static_feature_columns`` are specified, then the two sets of features
            are appended together.
        **kwargs
            Additional arguments. Same as of PandasDataset class.

        Returns
        -------
        PandasDataset
            Dataset containing series data from the given long dataframe.
        """
        if timestamp is not None:
            logger.info(f"Indexing data by '{timestamp}'.")
            dataframe.index = pd.to_datetime(dataframe[timestamp])

        if not isinstance(dataframe.index, DatetimeIndexOpsMixin):
            logger.info("Converting index into DatetimeIndex.")
            dataframe.index = pd.to_datetime(dataframe.index)

        if static_feature_columns is not None:
            logger.info(
                f"Collecting features from columns {static_feature_columns}."
            )
            other_static_features = (
                dataframe[[item_id] + static_feature_columns]
                .drop_duplicates()
                .set_index(item_id)
            )
            assert len(other_static_features) == len(
                dataframe[item_id].unique()
            )
        else:
            other_static_features = pd.DataFrame()

        logger.info(f"Grouping data by '{item_id}'; this may take some time.")
        pairs = list(dataframe.groupby(item_id))

        return cls(
            dataframes=pairs,
            static_features=pd.concat(
                [static_features, other_static_features], axis=1
            ),
            **kwargs,
        )


def pair_with_item_id(obj: Union[tuple, pd.DataFrame, pd.Series]):
    if isinstance(obj, tuple) and len(obj) == 2:
        return obj
    if isinstance(obj, (pd.DataFrame, pd.Series)):
        return (None, obj)
    raise ValueError("input must be a pair, or a pandas Series or DataFrame.")


def infer_freq(index: pd.Index) -> str:
    if isinstance(index, pd.PeriodIndex):
        return index.freqstr

    freq = pd.infer_freq(index)
    # pandas likes to infer the `start of x` frequency, however when doing
    # df.to_period("<x>S"), it fails, so we avoid using it. It's enough to
    # remove the trailing S, e.g `MS` -> `M
    if len(freq) > 1 and freq.endswith("S"):
        return freq[:-1]

    return freq


def is_uniform(index: pd.PeriodIndex) -> bool:
    """
    Check if ``index`` contains monotonically increasing periods, evenly spaced
    with frequency ``index.freq``.

        >>> ts = ["2021-01-01 00:00", "2021-01-01 02:00", "2021-01-01 04:00"]
        >>> is_uniform(pd.DatetimeIndex(ts).to_period("2H"))
        True
        >>> ts = ["2021-01-01 00:00", "2021-01-01 04:00"]
        >>> is_uniform(pd.DatetimeIndex(ts).to_period("2H"))
        False
    """

    return cast(bool, np.all(np.diff(index.asi8) == index.freq.n))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import math
from typing import List, Optional, Tuple

from mxnet import init
from mxnet.gluon import HybridBlock, nn, rnn

from gluonts.core.component import validated
from gluonts.mx import Tensor


class GatedLinearUnit(HybridBlock):
    @validated()
    def __init__(self, axis: int = -1, nonlinear: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.axis = axis
        self.nonlinear = nonlinear

    def hybrid_forward(self, F, x: Tensor) -> Tensor:
        val, gate = F.split(x, axis=self.axis, num_outputs=2)
        if self.nonlinear:
            val = F.tanh(val)
        gate = F.sigmoid(gate)
        return F.broadcast_mul(gate, val)


class GatedResidualNetwork(HybridBlock):
    @validated()
    def __init__(
        self,
        d_hidden: int,
        d_input: Optional[int] = None,
        d_output: Optional[int] = None,
        d_static: Optional[int] = None,
        dropout: float = 0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.d_hidden = d_hidden
        self.d_input = d_input or d_hidden
        self.d_static = d_static or 0
        if d_output is None:
            self.d_output = self.d_input
            self.add_skip = False
        else:
            self.d_output = d_output
            if d_output != self.d_input:
                self.add_skip = True
                with self.name_scope():
                    self.skip_proj = nn.Dense(
                        units=self.d_output,
                        in_units=self.d_input,
                        flatten=False,
                        weight_initializer=init.Xavier(),
                    )
            else:
                self.add_skip = False

        with self.name_scope():
            self.mlp = nn.HybridSequential(prefix="mlp_")
            self.mlp.add(
                nn.Dense(
                    units=self.d_hidden,
                    in_units=self.d_input + self.d_static,
                    flatten=False,
                    weight_initializer=init.Xavier(),
                )
            )
            self.mlp.add(nn.ELU())
            self.mlp.add(
                nn.Dense(
                    units=self.d_hidden,
                    in_units=self.d_hidden,
                    flatten=False,
                    weight_initializer=init.Xavier(),
                )
            )
            self.mlp.add(nn.Dropout(dropout)),
            self.mlp.add(
                nn.Dense(
                    units=self.d_output * 2,
                    in_units=self.d_hidden,
                    flatten=False,
                    weight_initializer=init.Xavier(),
                )
            )
            self.mlp.add(
                GatedLinearUnit(
                    axis=-1,
                    nonlinear=False,
                )
            )
            self.lnorm = nn.LayerNorm(axis=-1, in_channels=self.d_output)

    def hybrid_forward(
        self,
        F,
        x: Tensor,
        c: Optional[Tensor] = None,
    ) -> Tensor:
        if self.add_skip:
            skip = self.skip_proj(x)
        else:
            skip = x
        if self.d_static > 0 and c is None:
            raise ValueError("static variable is expected.")
        if self.d_static == 0 and c is not None:
            raise ValueError("static variable is not accpeted.")
        if c is not None:
            x = F.concat(x, c, dim=-1)
        x = self.mlp(x)
        x = self.lnorm(F.broadcast_add(x, skip))
        return x


class VariableSelectionNetwork(HybridBlock):
    @validated()
    def __init__(
        self,
        d_hidden: int,
        n_vars: int,
        dropout: float = 0.0,
        add_static: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.d_hidden = d_hidden
        self.n_vars = n_vars
        self.add_static = add_static

        with self.name_scope():
            self.weight_network = GatedResidualNetwork(
                d_hidden=self.d_hidden,
                d_input=self.d_hidden * self.n_vars,
                d_output=self.n_vars,
                d_static=self.d_hidden if add_static else None,
                dropout=dropout,
            )
            self.variable_network = []
            for n in range(self.n_vars):
                var_net = GatedResidualNetwork(
                    d_hidden=self.d_hidden,
                    dropout=dropout,
                )
                self.register_child(var_net, name=f"var_{n+1}")
                self.variable_network.append(var_net)

    def hybrid_forward(
        self,
        F,
        variables: List[Tensor],
        static: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Tensor]:
        if len(variables) != self.n_vars:
            raise ValueError(
                f"expect {self.n_vars} variables, {len(variables)} given."
            )
        if self.add_static and static is None:
            raise ValueError("static variable is expected.")
        if not self.add_static and static is not None:
            raise ValueError("static variable is not accpeted.")
        flatten = F.concat(*variables, dim=-1)
        if static is not None:
            static = F.broadcast_like(static, variables[0])
        weight = self.weight_network(flatten, static)
        weight = F.expand_dims(weight, axis=-2)
        weight = F.softmax(weight, axis=-1)
        var_encodings = []
        for var, net in zip(variables, self.variable_network):
            var_encodings.append(net(var))
        var_encodings = F.stack(*var_encodings, axis=-1)
        var_encodings = F.sum(F.broadcast_mul(var_encodings, weight), axis=-1)
        return var_encodings, weight


class SelfAttention(HybridBlock):
    @validated()
    def __init__(
        self,
        context_length: int,
        prediction_length: int,
        d_hidden: int,
        n_head: int = 1,
        bias: bool = True,
        share_values: bool = False,
        dropout: float = 0.0,
        temperature: float = 1.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if d_hidden % n_head != 0:
            raise ValueError(
                f"hidden dim {d_hidden} cannot be split into {n_head} heads."
            )
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.d_hidden = d_hidden
        self.n_head = n_head
        self.d_head = d_hidden // n_head
        self.bias = bias
        self.share_values = share_values
        self.temperature = temperature

        with self.name_scope():
            self.dropout = nn.Dropout(dropout)
            self.q_proj = nn.Dense(
                units=self.d_hidden,
                in_units=self.d_hidden,
                use_bias=self.bias,
                flatten=False,
                weight_initializer=init.Xavier(),
                prefix="q_proj_",
            )
            self.k_proj = nn.Dense(
                units=self.d_hidden,
                in_units=self.d_hidden,
                use_bias=self.bias,
                flatten=False,
                weight_initializer=init.Xavier(),
                prefix="k_proj_",
            )
            self.v_proj = nn.Dense(
                units=self.d_head if self.share_values else self.d_hidden,
                in_units=self.d_hidden,
                use_bias=self.bias,
                flatten=False,
                weight_initializer=init.Xavier(),
                prefix="v_proj_",
            )
            self.out_proj = nn.Dense(
                units=self.d_hidden,
                in_units=self.d_hidden,
                use_bias=self.bias,
                flatten=False,
                weight_initializer=init.Xavier(),
                prefix="out_proj_",
            )

    def _split_head(self, F, x: Tensor) -> Tensor:
        x = F.reshape(data=x, shape=(0, 0, -4, self.n_head, self.d_head))
        x = F.swapaxes(data=x, dim1=1, dim2=2)
        return x

    def _merge_head(self, F, x: Tensor) -> Tensor:
        x = F.swapaxes(data=x, dim1=1, dim2=2)
        x = F.reshape(data=x, shape=(0, 0, self.d_hidden))
        return x

    def _compute_qkv(self, F, x: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        cx = F.slice_axis(x, axis=1, begin=-self.prediction_length, end=None)
        q = self.q_proj(cx)
        k = self.k_proj(x)
        q = self._split_head(F, q)
        k = self._split_head(F, k)
        v = self.v_proj(x)
        if self.share_values:
            v = F.broadcast_like(v.expand_dims(axis=1), k)
        else:
            v = self._split_head(F, v)
        return q, k, v

    def _apply_mask(
        self, F, score: Tensor, key_mask: Optional[Tensor]
    ) -> Tensor:
        k_idx = F.contrib.arange_like(score, axis=-1)
        k_idx = (
            k_idx.expand_dims(axis=0).expand_dims(axis=0).expand_dims(axis=0)
        )
        q_idx = F.contrib.arange_like(score, axis=-2) + self.context_length
        q_idx = (
            q_idx.expand_dims(axis=-1).expand_dims(axis=0).expand_dims(axis=0)
        )
        unidir_mask = F.broadcast_lesser_equal(k_idx, q_idx)
        unidir_mask = F.broadcast_like(unidir_mask, score)
        score = F.where(unidir_mask, score, F.ones_like(score) * -1e9)
        if key_mask is not None:
            key_mask = key_mask.expand_dims(axis=1)  # head
            key_mask = key_mask.expand_dims(axis=2)  # query
            key_mask = F.broadcast_like(key_mask, score)
            score = F.where(key_mask, score, F.ones_like(score) * -1e9)
        return score

    def _compute_attn_score(
        self,
        F,
        q: Tensor,
        k: Tensor,
        mask: Optional[Tensor],
    ) -> Tensor:
        score = F.batch_dot(lhs=q, rhs=k, transpose_b=True)
        score = self._apply_mask(F, score, mask)
        score = score / (math.sqrt(self.d_head) * self.temperature)
        score = F.softmax(score, axis=-1)
        score = self.dropout(score)
        return score

    def _compute_attn_output(self, F, score: Tensor, v: Tensor) -> Tensor:
        v = F.batch_dot(score, v)
        v = self._merge_head(F, v)
        v = self.out_proj(v)
        return v

    def hybrid_forward(self, F, x: Tensor, mask: Optional[Tensor]) -> Tensor:
        q, k, v = self._compute_qkv(F, x)
        score = self._compute_attn_score(F, q, k, mask)
        v = self._compute_attn_output(F, score, v)
        return v


class TemporalFusionEncoder(HybridBlock):
    @validated()
    def __init__(
        self,
        context_length: int,
        prediction_length: int,
        d_input: int,
        d_hidden: int,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.context_length = context_length
        self.prediction_length = prediction_length
        with self.name_scope():
            self.encoder_lstm = rnn.HybridSequentialRNNCell(prefix="encoder_")
            self.encoder_lstm.add(
                rnn.LSTMCell(
                    hidden_size=d_hidden,
                    input_size=d_input,
                )
            )
            self.decoder_lstm = rnn.HybridSequentialRNNCell(prefix="decoder_")
            self.decoder_lstm.add(
                rnn.LSTMCell(
                    hidden_size=d_hidden,
                    input_size=d_input,
                )
            )
            self.gate = nn.HybridSequential()
            self.gate.add(
                nn.Dense(units=d_hidden * 2, in_units=d_hidden, flatten=False)
            )
            self.gate.add(GatedLinearUnit(axis=-1, nonlinear=False))
            if d_input != d_hidden:
                self.skip_proj = nn.Dense(
                    units=d_hidden, in_units=d_input, flatten=False
                )
                self.add_skip = True
            else:
                self.add_skip = False
            self.lnorm = nn.LayerNorm(axis=-1, in_channels=d_hidden)

    def hybrid_forward(
        self,
        F,
        ctx_input: Tensor,
        tgt_input: Tensor,
        states: List[Tensor],
    ) -> Tensor:
        ctx_encodings, states = self.encoder_lstm.unroll(
            length=self.context_length,
            inputs=ctx_input,
            begin_state=states,
            merge_outputs=True,
        )
        tgt_encodings, _ = self.decoder_lstm.unroll(
            length=self.prediction_length,
            inputs=tgt_input,
            begin_state=states,
            merge_outputs=True,
        )
        encodings = F.concat(ctx_encodings, tgt_encodings, dim=1)
        skip = F.concat(ctx_input, tgt_input, dim=1)
        if self.add_skip:
            skip = self.skip_proj(skip)
        encodings = self.gate(encodings)
        encodings = self.lnorm(F.broadcast_add(skip, encodings))
        return encodings


class TemporalFusionDecoder(HybridBlock):
    @validated()
    def __init__(
        self,
        context_length: int,
        prediction_length: int,
        d_hidden: int,
        d_var: int,
        n_head: int,
        dropout: float = 0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.context_length = context_length
        self.prediction_length = prediction_length

        with self.name_scope():
            self.enrich = GatedResidualNetwork(
                d_hidden=d_hidden,
                d_static=d_var,
                dropout=dropout,
            )
            self.attention = SelfAttention(
                context_length=context_length,
                prediction_length=prediction_length,
                d_hidden=d_hidden,
                n_head=n_head,
                share_values=True,
                dropout=dropout,
            )
            self.att_net = nn.HybridSequential(prefix="attention_")
            self.att_net.add(nn.Dropout(dropout))
            self.att_net.add(
                nn.Dense(
                    units=d_hidden * 2,
                    in_units=d_hidden,
                    flatten=False,
                    weight_initializer=init.Xavier(),
                )
            )
            self.att_net.add(
                GatedLinearUnit(
                    axis=-1,
                    nonlinear=False,
                )
            )
            self.att_lnorm = nn.LayerNorm(
                axis=-1,
                in_channels=d_hidden,
            )
            self.ff_net = nn.HybridSequential()
            self.ff_net.add(
                GatedResidualNetwork(
                    d_hidden,
                    dropout=dropout,
                )
            )
            self.ff_net.add(
                nn.Dense(
                    units=d_hidden * 2,
                    in_units=d_hidden,
                    flatten=False,
                    weight_initializer=init.Xavier(),
                )
            )
            self.ff_net.add(
                GatedLinearUnit(
                    axis=-1,
                    nonlinear=False,
                )
            )
            self.ff_lnorm = nn.LayerNorm(axis=-1, in_channels=d_hidden)

    def hybrid_forward(
        self, F, x: Tensor, static: Tensor, mask: Tensor
    ) -> Tensor:
        static = F.tile(
            static, reps=(1, self.context_length + self.prediction_length, 1)
        )
        skip = F.slice_axis(x, axis=1, begin=self.context_length, end=None)
        x = self.enrich(x, static)
        mask_pad = F.slice_axis(F.ones_like(mask), axis=1, begin=0, end=1)
        mask_pad = F.tile(mask_pad, reps=(1, self.prediction_length))
        mask = F.concat(mask, mask_pad, dim=1)
        att = self.attention(x, mask)
        att = self.att_net(att)
        x = F.slice_axis(x, axis=1, begin=self.context_length, end=None)
        x = self.att_lnorm(F.broadcast_add(x, att))
        x = self.ff_net(x)
        x = self.ff_lnorm(F.broadcast_add(x, skip))
        return x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import itertools
import pickle
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Tuple

import numpy as np
from numpy.testing import assert_equal
import pytest

from gluonts.dataset.artificial import constant_dataset
from gluonts.itertools import (
    batcher,
    Cached,
    PickleCached,
    Cyclic,
    IterableSlice,
    SizedIterableSlice,
    PseudoShuffled,
    rows_to_columns,
    columns_to_rows,
    select,
    pluck_attr,
    power_set,
    Map,
    StarMap,
    Filter,
    join_items,
)


@pytest.mark.parametrize(
    "data, n, expected", [([1, 2, 3], 7, [1, 2, 3, 1, 2, 3, 1]), ([], 4, [])]
)
def test_cyclic(data: Iterable, n: int, expected: List) -> None:
    cyclic_data = Cyclic(data)
    actual = list(itertools.islice(cyclic_data, n))
    assert actual == expected


@pytest.mark.parametrize(
    "data",
    [
        range(20),
        constant_dataset()[1],
    ],
)
def test_pseudo_shuffled(data: Iterable) -> None:
    list_data = list(data)
    shuffled_iter = PseudoShuffled(iter(list_data), shuffle_buffer_length=5)
    shuffled_data = list(shuffled_iter)
    assert len(shuffled_data) == len(list_data)
    assert all(d in shuffled_data for d in list_data)


@pytest.mark.parametrize(
    "data, expected_elements_per_iteration",
    [
        (Cached(range(4)), (list(range(4)),) * 5),
        (PickleCached(range(4)), (list(range(4)),) * 5),
        (batcher(range(10), 3), ([[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]], [])),
        (IterableSlice(range(10), 3), ([0, 1, 2],) * 5),
        (
            IterableSlice(iter(range(10)), 3),
            ([0, 1, 2], [3, 4, 5], [6, 7, 8], [9], []),
        ),
        (
            IterableSlice(iter(Cyclic(range(5))), 3),
            ([0, 1, 2], [3, 4, 0], [1, 2, 3], [4, 0, 1]),
        ),
    ],
)
def test_iterate_multiple_times(
    data: Iterable, expected_elements_per_iteration: Tuple[List]
):
    for expected_elements in expected_elements_per_iteration:
        assert list(data) == expected_elements


@pytest.mark.parametrize(
    "iterable, assert_content",
    [
        (Cached(range(5)), True),
        (PseudoShuffled(range(20), 5), False),
        (IterableSlice(Cyclic(range(5)), 9), True),
    ],
)
def test_pickle(iterable: Iterable, assert_content: bool):
    with tempfile.TemporaryDirectory() as tmpdir:
        with open(Path(tmpdir) / "temp.pickle", "wb") as fp:
            pickle.dump(iterable, fp)

        with open(Path(tmpdir) / "temp.pickle", "rb") as fp:
            iterable_copy = pickle.load(fp)

    data = list(iterable)
    data_copy = list(iterable_copy)

    assert len(data) == len(data_copy)

    if assert_content:
        assert data == data_copy


@pytest.mark.parametrize(
    "given, expected",
    [
        ([], {}),
        ([{"a": 1, "b": 2}, {"a": 3, "b": 4}], {"a": [1, 3], "b": [2, 4]}),
    ],
)
@pytest.mark.parametrize("wrapper", [list, np.array])
def test_rows_to_columns(given, expected, wrapper):
    output = rows_to_columns(given, wrapper)
    assert_equal(output, expected)
    assert columns_to_rows(output) == given


def test_sized_iterable_slice():
    def generator():
        for i in range(10):
            yield i

    unsized_iter = generator()
    sized_iter = list(range(10))

    unsized_iter_slice = SizedIterableSlice(unsized_iter, 5)
    with pytest.raises(TypeError):
        len(unsized_iter_slice)

    sized_iter_slice = SizedIterableSlice(sized_iter, 5)
    assert len(sized_iter_slice) == 5

    sized_iter_slice = SizedIterableSlice(sized_iter, 15)
    assert len(sized_iter_slice) == 10


def test_select():
    d = {"a": 1, "b": 2, "c": 3}

    assert select("abc", d) == d
    assert select("ab", d) == {"a": 1, "b": 2}

    assert select("abd", d, ignore_missing=True) == {"a": 1, "b": 2}

    with pytest.raises(KeyError):
        select("abd", d, ignore_missing=False) == {"a": 1, "b": 2}


def test_map():
    data = [1, 2, 3]
    applied = Map(lambda n: n + 1, data)

    assert list(applied) == [2, 3, 4]


def test_starmap():
    def add(a, b, c):
        return a + b + c

    data = [[1, 2, 3], [4, 5, 6]]
    applied = StarMap(add, data)

    assert list(applied) == [add(1, 2, 3), add(4, 5, 6)]


def test_filter():
    data = [1, 2, 3]
    applied = Filter(lambda n: n <= 2, data)

    assert list(applied) == [1, 2]


def test_pluck_attr():
    @dataclass
    class X:
        a: int
        b: int

    xs = [X(1, 2), X(3, 4)]

    assert pluck_attr(xs, "a") == [1, 3]
    assert pluck_attr(xs, "b") == [2, 4]

    with pytest.raises(AttributeError):
        assert pluck_attr(xs, "c")

    assert pluck_attr(xs, "c", 4) == [4, 4]


def test_pluck_attr_curry():
    @dataclass
    class X:
        a: int
        b: int

    xs = [X(1, 2), X(3, 4)]

    get = pluck_attr(xs)

    assert get("a") == [1, 3]
    assert get("b") == [2, 4]

    with pytest.raises(AttributeError):
        assert get("c")

    assert get("c", 4) == [4, 4]


def test_power_set():
    collection = list(range(4))
    subsets = list(power_set(collection))

    expected_subsets = [
        (),
        (0,),
        (1,),
        (2,),
        (3,),
        (0, 1),
        (0, 2),
        (0, 3),
        (1, 2),
        (1, 3),
        (2, 3),
        (0, 1, 2),
        (0, 1, 3),
        (0, 2, 3),
        (1, 2, 3),
        (0, 1, 2, 3),
    ]

    assert len(subsets) == 2 ** len(collection)
    assert len(expected_subsets) == 2 ** len(collection)

    for es in expected_subsets:
        assert es in subsets


def test_join_items():
    left = {"a": 1, "b": 2}
    right = {"a": 3, "c": 4}

    assert list(join_items(left, right)) == [
        ("a", 1, 3),
        ("b", 2, None),
        ("c", None, 4),
    ]

    assert list(join_items(left, right, "inner")) == [
        ("a", 1, 3),
    ]

    assert list(join_items(left, right, "left")) == [
        ("a", 1, 3),
        ("b", 2, None),
    ]

    assert list(join_items(left, right, "right")) == [
        ("a", 1, 3),
        ("c", None, 4),
    ]

    with pytest.raises(Exception):
        oin_items(left, right, "strict")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Tuple

import mxnet as mx
import numpy as np
import pytest

from gluonts.core.serde import dump_json, load_json
from gluonts.mx.distribution import (
    FixedKnotsPiecewiseLinearOutput,
    PiecewiseLinear,
    PiecewiseLinearOutput,
)
from gluonts.testutil import empirical_cdf

serialize_fn_list = [lambda x: x, lambda x: load_json(dump_json(x))]


@pytest.mark.parametrize(
    "distr, target, expected_target_cdf, expected_target_crps",
    [
        (
            PiecewiseLinear(
                gamma=mx.nd.ones(shape=(1,)),
                slopes=mx.nd.array([2, 3, 1]).reshape(shape=(1, 3)),
                knot_spacings=mx.nd.array([0.3, 0.4, 0.3]).reshape(
                    shape=(1, 3)
                ),
            ),
            [2.2],
            [0.5],
            [0.223000],
        ),
        (
            PiecewiseLinear(
                gamma=mx.nd.ones(shape=(2,)),
                slopes=mx.nd.array([[1, 1], [1, 2]]).reshape(shape=(2, 2)),
                knot_spacings=mx.nd.array([[0.4, 0.6], [0.4, 0.6]]).reshape(
                    shape=(2, 2)
                ),
            ),
            [1.5, 1.6],
            [0.5, 0.5],
            [0.083333, 0.145333],
        ),
    ],
)
@pytest.mark.parametrize("serialize_fn", serialize_fn_list)
def test_values(
    distr: PiecewiseLinear,
    target: List[float],
    expected_target_cdf: List[float],
    expected_target_crps: List[float],
    serialize_fn,
):
    distr = serialize_fn(distr)
    target = mx.nd.array(target).reshape(shape=(len(target),))
    expected_target_cdf = np.array(expected_target_cdf).reshape(
        (len(expected_target_cdf),)
    )
    expected_target_crps = np.array(expected_target_crps).reshape(
        (len(expected_target_crps),)
    )

    assert all(np.isclose(distr.cdf(target).asnumpy(), expected_target_cdf))
    assert all(np.isclose(distr.crps(target).asnumpy(), expected_target_crps))

    # compare with empirical cdf from samples
    num_samples = 100_000
    samples = distr.sample(num_samples).asnumpy()
    assert np.isfinite(samples).all()

    emp_cdf, edges = empirical_cdf(samples)
    calc_cdf = distr.cdf(mx.nd.array(edges)).asnumpy()
    assert np.allclose(calc_cdf[1:, :], emp_cdf, atol=1e-2)


@pytest.mark.parametrize(
    "batch_shape, num_pieces, num_samples",
    [((3, 4, 5), 10, 100), ((1,), 2, 1), ((10,), 10, 10), ((10, 5), 2, 1)],
)
@pytest.mark.parametrize("serialize_fn", serialize_fn_list)
def test_shapes(
    batch_shape: Tuple, num_pieces: int, num_samples: int, serialize_fn
):
    gamma = mx.nd.ones(shape=(*batch_shape,))
    slopes = mx.nd.ones(shape=(*batch_shape, num_pieces))  # all positive
    knot_spacings = (
        mx.nd.ones(shape=(*batch_shape, num_pieces)) / num_pieces
    )  # positive and sum to 1
    target = mx.nd.ones(shape=batch_shape)  # shape of gamma

    distr = PiecewiseLinear(
        gamma=gamma, slopes=slopes, knot_spacings=knot_spacings
    )
    distr = serialize_fn(distr)

    # assert that the parameters and target have proper shapes
    assert gamma.shape == target.shape
    assert knot_spacings.shape == slopes.shape
    assert len(gamma.shape) + 1 == len(knot_spacings.shape)

    # assert that batch_shape is computed properly
    assert distr.batch_shape == batch_shape

    # assert that shapes of original parameters are correct
    assert distr.b.shape == slopes.shape
    assert distr.knot_positions.shape == knot_spacings.shape

    # assert that the shape of crps is correct
    assert distr.crps(target).shape == batch_shape

    # assert that the quantile shape is correct when computing the quantile values at knot positions - used for a_tilde
    assert distr.quantile_internal(knot_spacings, axis=-2).shape == (
        *batch_shape,
        num_pieces,
    )

    # assert that the samples and the quantile values shape when num_samples is None is correct
    samples = distr.sample()
    assert samples.shape == batch_shape
    assert distr.quantile_internal(samples).shape == batch_shape

    # assert that the samples and the quantile values shape when num_samples is not None is correct
    samples = distr.sample(num_samples)
    assert samples.shape == (num_samples, *batch_shape)
    assert distr.quantile_internal(samples, axis=0).shape == (
        num_samples,
        *batch_shape,
    )


def test_simple_symmetric():
    gamma = mx.nd.array([-1.0])
    slopes = mx.nd.array([[2.0, 2.0]])
    knot_spacings = mx.nd.array([[0.5, 0.5]])

    distr = PiecewiseLinear(
        gamma=gamma, slopes=slopes, knot_spacings=knot_spacings
    )

    assert distr.cdf(mx.nd.array([-2.0])).asnumpy().item() == 0.0
    assert distr.cdf(mx.nd.array([+2.0])).asnumpy().item() == 1.0

    expected_crps = np.array([1.0 + 2.0 / 3.0])

    assert np.allclose(
        distr.crps(mx.nd.array([-2.0])).asnumpy(), expected_crps
    )

    assert np.allclose(distr.crps(mx.nd.array([2.0])).asnumpy(), expected_crps)


def test_robustness():
    distr_out = PiecewiseLinearOutput(num_pieces=10)
    args_proj = distr_out.get_args_proj()
    args_proj.initialize()

    net_out = mx.nd.random.normal(shape=(1000, 30), scale=1e2)
    gamma, slopes, knot_spacings = args_proj(net_out)
    distr = distr_out.distribution((gamma, slopes, knot_spacings))

    # compute the 1-quantile (the 0-quantile is gamma)
    sup_support = gamma + mx.nd.sum(slopes * knot_spacings, axis=-1)

    assert mx.nd.broadcast_lesser_equal(gamma, sup_support).asnumpy().all()

    width = sup_support - gamma
    x = mx.random.uniform(low=gamma - width, high=sup_support + width)

    # check that 0 < cdf < 1
    cdf_x = distr.cdf(x)
    assert (
        mx.nd.min(cdf_x).asscalar() >= 0.0
        and mx.nd.max(cdf_x).asscalar() <= 1.0
    )

    # check that 0 <= crps
    crps_x = distr.crps(x)
    assert mx.nd.min(crps_x).asscalar() >= 0.0


def test_fkpwl_distr_output_same_as_pl():
    x = mx.nd.array([[0.1, 0.5], [0.3, 0.99], [0.2, 0.6]])

    gamma = mx.nd.array([0.0, 2.0])
    slopes = mx.nd.array([[0.1, 0.5, 0.9], [2.0, 0.1, 5.0]])
    knot_spacings = mx.nd.array([[0.1, 0.5, 0.4], [0.1, 0.5, 0.4]])

    pl_output = PiecewiseLinearOutput(num_pieces=3)
    pl_dist = pl_output.distribution([gamma, slopes, knot_spacings])

    fkpl = FixedKnotsPiecewiseLinearOutput(
        [0.1, 0.6],
    ).distribution([gamma, slopes, knot_spacings])

    assert np.allclose(
        pl_dist.crps(x).asnumpy(),
        fkpl.crps(x).asnumpy(),
    )


def test_fkpwl_distr_args_correct():
    x = mx.nd.array([[0.1, 0.5], [0.3, 0.99], [0.2, 0.6]])

    knot_spacings = mx.nd.array(
        [[0.1, 0.5, 0.4], [0.1, 0.5, 0.4], [0.1, 0.5, 0.4]]
    )

    fkpl_proj = FixedKnotsPiecewiseLinearOutput(
        [0.1, 0.6],
    ).get_args_proj()

    fkpl_proj.initialize()
    _, _, ks = fkpl_proj(x)

    assert np.allclose(ks.asnumpy(), knot_spacings.asnumpy())



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np


class WeightedSampler:
    """
    Utility class for sampling indices based on unnormalized weights.
    """

    @staticmethod
    def sample(weights, num_samples):
        """
        Sample indices according to `weights`: `ix` is chosen with probability
        `weights`[`ix`]

        `weights` need not sum to 1.

        :param weights:
        :param num_samples:
        :return:
        """
        assert all(weights >= 0.0), "Sampling weights must be non-negative"
        # In the special case where all the weights are zeros, we want to
        # sample all indices uniformly
        weights = np.ones_like(weights) if sum(weights) == 0.0 else weights

        cumsum_weights = np.cumsum(weights)

        # Just for better readability.
        total_weight = cumsum_weights[-1]

        # Samples from the Uniform distribution: U(0, `total_weight`)
        uniform_samples = total_weight * np.random.random(num_samples)

        # Search for the last `ix` for each sample u ~ U(0, total weight)
        # such that u <= `cumsum`[`ix`]
        # This means `ix` is chosen with probability
        # `cumsum`[`ix`] - `cumsum`[`ix` - 1] = weights[ix]
        samples_ix = np.searchsorted(
            cumsum_weights, uniform_samples, side="left"
        )

        return samples_ix



from typing import Callable, Optional, TypeVar, Generic

T = TypeVar("T")
U = TypeVar("U")

def expect(val: Optional[T], msg: str) -> T: ...
def do(val: Optional[T], fn: Callable[[T], U]) -> Optional[T]: ...
def map(val: Optional[T], fn: Callable[[T], U]) -> Optional[U]: ...
def map_or(val: Optional[T], fn: Callable[[T], U], default: U) -> U: ...
def map_or_else(
    val: Optional[T], fn: Callable[[T], U], factory: Callable[[], U]
) -> U: ...
def unwrap(val: Optional[T]) -> T: ...
def unwrap_or(val: Optional[T], default: T) -> T: ...
def unwrap_or_else(val: Optional[T], factory: Callable[[], T]) -> T: ...
def and_(val: Optional[T], other: Optional[U]) -> Optional[U]: ...
def and_then(
    val: Optional[T], fn: Callable[[T], Optional[U]]
) -> Optional[U]: ...
def or_(val: Optional[T], default: Optional[T]) -> Optional[T]: ...
def or_else(
    val: Optional[T], factory: Callable[[], Optional[T]]
) -> Optional[T]: ...
def contains(val: Optional[T], other: U) -> bool: ...
def filter(val: Optional[T], pred: Callable[[T], bool]) -> Optional[T]: ...
def xor(val: Optional[T], other: Optional[T]) -> Optional[T]: ...
def flatten(val: Optional[Optional[T]]) -> Optional[T]: ...

class Maybe(Generic[T]):
    val: Optional[T]
    def expect(self, msg: str) -> T: ...
    def do(self, fn: Callable[[T], U]) -> Optional[T]: ...
    def map(self, fn: Callable[[T], U]) -> Optional[U]: ...
    def map_or(self, fn: Callable[[T], U], default: U) -> U: ...
    def map_or_else(
        self, fn: Callable[[T], U], factory: Callable[[], U]
    ) -> U: ...
    def unwrap(self) -> T: ...
    def unwrap_or(self, default: T) -> T: ...
    def unwrap_or_else(self, fn: Callable[[], T]) -> T: ...
    def and_(self, other: Optional[U]) -> Optional[U]: ...
    def __and__(self, other: Optional[U]) -> Optional[U]: ...
    def and_then(self, fn: Callable[[T], Optional[U]]) -> Optional[U]: ...
    def or_(self, default: Optional[T]) -> Optional[T]: ...
    def __or__(self, default: Optional[T]) -> Optional[T]: ...
    def or_else(self, factory: Callable[[], Optional[T]]) -> Optional[T]: ...
    def contains(self, other: U) -> bool: ...
    def filter(self, pred: Callable[[T], bool]) -> Optional[T]: ...
    def xor(self, other: Optional[T]) -> Optional[T]: ...
    def __xor__(self, other: Optional[T]) -> Optional[T]: ...
    def flatten(self) -> Optional[T]: ...
    def __init__(self, val) -> None: ...



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from gluonts.core.serde import encode, decode, dump_json
from gluonts.mx.model.tpp.distribution import (
    LoglogisticOutput,
    WeibullOutput,
)
from gluonts.mx.distribution import (
    BetaOutput,
    CategoricalOutput,
    DeterministicOutput,
    DistributionOutput,
    DirichletMultinomialOutput,
    DirichletOutput,
    EmpiricalDistributionOutput,
    GammaOutput,
    GaussianOutput,
    GenParetoOutput,
    LaplaceOutput,
    LogitNormalOutput,
    LowrankMultivariateGaussianOutput,
    MultivariateGaussianOutput,
    NegativeBinomialOutput,
    OneInflatedBetaOutput,
    PiecewiseLinearOutput,
    PoissonOutput,
    StudentTOutput,
    UniformOutput,
    ZeroAndOneInflatedBetaOutput,
    ZeroInflatedBetaOutput,
    ZeroInflatedNegativeBinomialOutput,
    ZeroInflatedPoissonOutput,
)


@pytest.mark.parametrize(
    "distr_output",
    [
        BetaOutput(),
        CategoricalOutput(num_cats=3),
        DeterministicOutput(value=42.0),
        DirichletMultinomialOutput(dim=3, n_trials=5),
        DirichletOutput(dim=4),
        EmpiricalDistributionOutput(
            num_samples=10, distr_output=GaussianOutput()
        ),
        GammaOutput(),
        GaussianOutput(),
        GenParetoOutput(),
        LaplaceOutput(),
        LogitNormalOutput(),
        LoglogisticOutput(),
        LowrankMultivariateGaussianOutput(dim=5, rank=2),
        MultivariateGaussianOutput(dim=4),
        NegativeBinomialOutput(),
        OneInflatedBetaOutput(),
        PiecewiseLinearOutput(num_pieces=10),
        PoissonOutput(),
        StudentTOutput(),
        UniformOutput(),
        WeibullOutput(),
        ZeroAndOneInflatedBetaOutput(),
        ZeroInflatedBetaOutput(),
        ZeroInflatedNegativeBinomialOutput(),
        ZeroInflatedPoissonOutput(),
    ],
)
def test_distribution_output_serde(distr_output: DistributionOutput):
    distr_output_copy = decode(encode(distr_output))

    assert isinstance(distr_output_copy, type(distr_output))
    assert dump_json(distr_output_copy) == dump_json(distr_output)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx
import numpy as np
import pytest

import gluonts.mx.distribution.bijection as bij
from gluonts.mx.distribution import (
    Beta,
    Binned,
    Dirichlet,
    Gamma,
    Laplace,
    MixtureDistribution,
    NegativeBinomial,
    PiecewiseLinear,
    Poisson,
    StudentT,
    TransformedDistribution,
    Uniform,
)
from gluonts.mx.distribution.box_cox_transform import BoxCoxTransform
from gluonts.mx.distribution.gaussian import Gaussian


@pytest.mark.parametrize(
    "slice_axis_args, expected_axis_length",
    [[(0, 0, None), 3], [(0, 1, 3), 2], [(1, -1, None), 1]],
)
@pytest.mark.parametrize(
    "distr",
    [
        Gaussian(
            mu=mx.nd.random.normal(shape=(3, 4)),
            sigma=mx.nd.random.uniform(shape=(3, 4)),
        )
    ],
)
def test_distr_slice_axis(distr, slice_axis_args, expected_axis_length):
    axis, begin, end = slice_axis_args
    distr_sliced = distr.slice_axis(axis, begin, end)

    assert distr_sliced.batch_shape[axis] == expected_axis_length


class SliceHelper:
    def __getitem__(self, item):
        return item


sh = SliceHelper()

BATCH_SHAPE = (3, 4, 5)


DISTRIBUTIONS_WITH_QUANTILE_FUNCTION = (Gaussian, Uniform, Laplace, Binned)


@pytest.mark.parametrize(
    "distr",
    [
        TransformedDistribution(
            Gaussian(
                mu=mx.nd.random.uniform(shape=BATCH_SHAPE),
                sigma=mx.nd.ones(shape=BATCH_SHAPE),
            ),
            [
                bij.AffineTransformation(
                    scale=1e-1 + mx.nd.random.uniform(shape=BATCH_SHAPE)
                ),
                bij.softrelu,
            ],
        ),
        Binned(
            bin_log_probs=mx.nd.uniform(shape=BATCH_SHAPE + (23,)),
            bin_centers=mx.nd.array(np.logspace(-1, 1, 23))
            + mx.nd.zeros(BATCH_SHAPE + (23,)),
        ),
        TransformedDistribution(
            Binned(
                bin_log_probs=mx.nd.uniform(shape=BATCH_SHAPE + (23,)),
                bin_centers=mx.nd.array(np.logspace(-1, 1, 23))
                + mx.nd.zeros(BATCH_SHAPE + (23,)),
            ),
            [
                bij.AffineTransformation(
                    scale=1e-1 + mx.nd.random.uniform(shape=BATCH_SHAPE)
                ),
                bij.softrelu,
            ],
        ),
        Gaussian(
            mu=mx.nd.zeros(shape=BATCH_SHAPE),
            sigma=mx.nd.ones(shape=BATCH_SHAPE),
        ),
        Gamma(
            alpha=mx.nd.ones(shape=BATCH_SHAPE),
            beta=mx.nd.ones(shape=BATCH_SHAPE),
        ),
        Beta(
            alpha=0.5 * mx.nd.ones(shape=BATCH_SHAPE),
            beta=0.5 * mx.nd.ones(shape=BATCH_SHAPE),
        ),
        StudentT(
            mu=mx.nd.zeros(shape=BATCH_SHAPE),
            sigma=mx.nd.ones(shape=BATCH_SHAPE),
            nu=mx.nd.ones(shape=BATCH_SHAPE),
        ),
        Dirichlet(alpha=mx.nd.ones(shape=BATCH_SHAPE)),
        Laplace(
            mu=mx.nd.zeros(shape=BATCH_SHAPE), b=mx.nd.ones(shape=BATCH_SHAPE)
        ),
        NegativeBinomial(
            mu=mx.nd.zeros(shape=BATCH_SHAPE),
            alpha=mx.nd.ones(shape=BATCH_SHAPE),
        ),
        Poisson(rate=mx.nd.ones(shape=BATCH_SHAPE)),
        Uniform(
            low=-mx.nd.ones(shape=BATCH_SHAPE),
            high=mx.nd.ones(shape=BATCH_SHAPE),
        ),
        PiecewiseLinear(
            gamma=mx.nd.ones(shape=BATCH_SHAPE),
            slopes=mx.nd.ones(shape=(3, 4, 5, 10)),
            knot_spacings=mx.nd.ones(shape=(3, 4, 5, 10)) / 10,
        ),
        MixtureDistribution(
            mixture_probs=mx.nd.stack(
                0.2 * mx.nd.ones(shape=BATCH_SHAPE),
                0.8 * mx.nd.ones(shape=BATCH_SHAPE),
                axis=-1,
            ),
            components=[
                Gaussian(
                    mu=mx.nd.zeros(shape=BATCH_SHAPE),
                    sigma=mx.nd.ones(shape=BATCH_SHAPE),
                ),
                StudentT(
                    mu=mx.nd.zeros(shape=BATCH_SHAPE),
                    sigma=mx.nd.ones(shape=BATCH_SHAPE),
                    nu=mx.nd.ones(shape=BATCH_SHAPE),
                ),
            ],
        ),
        TransformedDistribution(
            StudentT(
                mu=mx.nd.zeros(shape=BATCH_SHAPE),
                sigma=mx.nd.ones(shape=BATCH_SHAPE),
                nu=mx.nd.ones(shape=BATCH_SHAPE),
            ),
            [
                bij.AffineTransformation(
                    scale=1e-1 + mx.nd.random.uniform(shape=BATCH_SHAPE)
                )
            ],
        ),
        TransformedDistribution(
            Uniform(
                low=mx.nd.zeros(shape=BATCH_SHAPE),
                high=mx.nd.ones(shape=BATCH_SHAPE),
            ),
            [
                BoxCoxTransform(
                    lambda_1=mx.nd.ones(shape=BATCH_SHAPE),
                    lambda_2=mx.nd.zeros(shape=BATCH_SHAPE),
                )
            ],
        ),
    ],
)
@pytest.mark.parametrize(
    "slice_item", [sh[1:2], sh[1, :], sh[:, 0], sh[0, -1]]
)
def test_slice_axis_results(distr, slice_item):
    s = distr.sample().asnumpy()
    sliced = distr[slice_item]
    s_sliced = sliced.sample().asnumpy()
    assert s_sliced.shape == s[slice_item].shape

    y = np.random.uniform(size=BATCH_SHAPE)
    lp_expected = distr.loss(mx.nd.array(y)).asnumpy()[slice_item]
    lp_actual = sliced.loss(mx.nd.array(y[slice_item])).asnumpy()
    assert np.allclose(lp_actual, lp_expected)

    tmp = (
        distr.base_distribution
        if isinstance(distr, TransformedDistribution)
        else distr
    )
    has_quantile_fn = isinstance(tmp, DISTRIBUTIONS_WITH_QUANTILE_FUNCTION)

    if has_quantile_fn:
        for ql in [0.01, 0.1, 0.5, 0.9, 0.99]:
            qs_actual = sliced.quantile(mx.nd.array([ql])).asnumpy()[0]
            qs_expected = distr.quantile(mx.nd.array([ql])).asnumpy()[0][
                slice_item
            ]
            assert np.allclose(qs_actual, qs_expected)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict

import numpy as np


def absolute_label(data: Dict[str, np.ndarray]) -> np.ndarray:
    return np.abs(data["label"])


def error(data: Dict[str, np.ndarray], forecast_type: str) -> np.ndarray:
    return data["label"] - data[forecast_type]


def absolute_error(
    data: Dict[str, np.ndarray], forecast_type: str
) -> np.ndarray:
    return np.abs(error(data, forecast_type))


def squared_error(
    data: Dict[str, np.ndarray], forecast_type: str
) -> np.ndarray:
    return np.square(error(data, forecast_type))


def quantile_loss(data: Dict[str, np.ndarray], q: float) -> np.ndarray:
    forecast_type = str(q)
    prediction = data[forecast_type]

    return 2 * np.abs(
        error(data, forecast_type) * ((prediction >= data["label"]) - q)
    )


def coverage(data: Dict[str, np.ndarray], q: float) -> np.ndarray:
    forecast_type = str(q)
    return (data["label"] <= data[forecast_type]).astype(float)


def absolute_percentage_error(
    data: Dict[str, np.ndarray], forecast_type: str
) -> np.ndarray:
    return absolute_error(data, forecast_type) / absolute_label(data)


def symmetric_absolute_percentage_error(
    data: Dict[str, np.ndarray], forecast_type: str
) -> np.ndarray:
    return (
        2
        * absolute_error(data, forecast_type)
        / (absolute_label(data) + np.abs(data[forecast_type]))
    )


def scaled_interval_score(
    data: Dict[str, np.ndarray], alpha: float
) -> np.ndarray:
    lower_quantile = data[str(alpha / 2)]
    upper_quantile = data[str(1.0 - alpha / 2)]
    label = data["label"]

    numerator = (
        upper_quantile
        - lower_quantile
        + 2.0 / alpha * (lower_quantile - label) * (label < lower_quantile)
        + 2.0 / alpha * (label - upper_quantile) * (label > upper_quantile)
    )

    return numerator / data["seasonal_error"]


def absolute_scaled_error(
    data: Dict[str, np.ndarray], forecast_type: str
) -> np.ndarray:
    return absolute_error(data, forecast_type) / data["seasonal_error"]
